import functools
from abc import ABC, abstractmethod
from enum import Enum, unique
from typing import Annotated, Any, Callable, Mapping, ParamSpec, Protocol, get_args

from ordered_set import OrderedSet, OrderedSetInitializer
from pydantic import (
    AfterValidator,
    GetCoreSchemaHandler,
    GetPydanticSchema,
    HttpUrl,
)
from pydantic_core import core_schema
from typing_extensions import TypeVar
from yarl import URL

_T = TypeVar("_T")

_P = ParamSpec("_P")
_R = TypeVar("_R")

_T_float = TypeVar("_T_float", default=float)


Vec2 = tuple[_T_float, _T_float]
Vec3 = tuple[_T_float, _T_float, _T_float]
Vec4 = tuple[_T_float, _T_float, _T_float, _T_float]


class Sortable(ABC):
    """ABC for classes which can be sorted."""

    @property
    @abstractmethod
    def _cmp_key(self) -> Any: ...

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Sortable):
            return self._cmp_key < other._cmp_key
        return NotImplemented


_T_Sortable = TypeVar("_T_Sortable", bound=Sortable)

_T_covariant = TypeVar("_T_covariant", covariant=True)


def sorted_dict(d: Mapping[_T, _T_Sortable]) -> dict[_T, _T_Sortable]:
    return dict(sorted(d.items(), key=lambda item: item[1]))


class IProperty(Protocol[_T_covariant]):
    def __get__(
        self,
        __instance: Any,
        __owner: type | None = None,
        /,
    ) -> _T_covariant: ...


FieldOrProperty = _T_covariant | IProperty[_T_covariant]


@unique
class TryGetEnum(Enum):
    @classmethod
    def get(cls, value: Any):
        try:
            return cls(value)
        except ValueError:
            return None


# https://docs.pydantic.dev/latest/concepts/types/#generic-containers
class PydanticOrderedSet(OrderedSet[_T]):
    def __init__(self, initial: OrderedSetInitializer[_T] | None = None):
        super().__init__(initial or [])

    @classmethod
    def collect(cls, *initial: _T):
        return cls(initial)

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        type_arg: Any
        match len(type_args := get_args(source)):
            case 0:
                type_arg = Any
            case 1:
                type_arg = type_args[0]
            case n:
                raise ValueError(f"Expected 0 or 1 type args, got {n}: {type_args}")

        return core_schema.union_schema(
            [
                core_schema.is_instance_schema(cls),
                cls._get_non_instance_schema(type_arg, handler),
            ],
            serialization=cls._get_ser_schema(type_arg, handler),
        )

    @classmethod
    def _get_non_instance_schema(
        cls,
        type_arg: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        # validate from OrderedSetInitializer
        return core_schema.no_info_after_validator_function(
            function=PydanticOrderedSet,
            schema=handler.generate_schema(OrderedSetInitializer[type_arg]),
        )

    @classmethod
    def _get_ser_schema(
        cls,
        type_arg: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.SerSchema:
        # serialize to list
        return core_schema.plain_serializer_function_ser_schema(
            function=cls._get_items,
        )

    def _get_items(self):
        return self.items


PydanticURL = Annotated[
    URL,
    GetPydanticSchema(
        lambda typ, handler: core_schema.union_schema(
            [
                core_schema.is_instance_schema(typ),
                core_schema.no_info_after_validator_function(
                    function=lambda raw: URL(raw.rstrip("/")),
                    schema=handler.generate_schema(Annotated[str, HttpUrl]),
                ),
            ],
            serialization=core_schema.plain_serializer_function_ser_schema(
                function=lambda url: str(url).rstrip("/"),
            ),
        )
    ),
]


def clamping_validator(lower: float | None, upper: float | None):
    def validator(value: float):
        lower_ = lower if lower is not None else value
        upper_ = upper if upper is not None else value
        return max(lower_, min(upper_, value))

    return AfterValidator(validator)


# short alias for convenience
clamped = clamping_validator


def typed_partial(f: Callable[_P, _R]) -> Callable[_P, Callable[_P, _R]]:
    """Given a function, returns a function that takes arguments for that function and
    returns a function that calls the original function with the partial arguments and
    whatever's passed into it.

    Basically, this is a more strongly typed version of `functools.partial`.
    """

    @functools.wraps(f)
    def builder_builder(*partial_args: _P.args, **partial_kwargs: _P.kwargs):
        @functools.wraps(f)
        def builder(*args: _P.args, **kwargs: _P.kwargs):
            return f(*partial_args, *args, **partial_kwargs, **kwargs)

        return builder

    return builder_builder


def cast_nullable(value: _T) -> _T | None:
    """Tells the type checker that the given value could also be `None`.

    At runtime, just returns the value as-is.
    """
    return value



from pathlib import Path

from hexdoc.minecraft.assets import Texture, TextureLookups
from hexdoc.model import HexdocModel
from hexdoc.utils.types import PydanticURL


class HexdocMetadata(HexdocModel):
    """Automatically generated at `export_dir/modid.hexdoc.json`."""

    book_url: PydanticURL | None
    """Github Pages base url."""
    asset_url: PydanticURL
    """raw.githubusercontent.com base url."""
    textures: TextureLookups[Texture]

    @classmethod
    def path(cls, modid: str) -> Path:
        return Path(f"{modid}.hexdoc.json")


def load_metadata_textures(all_metadata: dict[str, HexdocMetadata]):
    lookups = TextureLookups[Texture](dict)

    for metadata in all_metadata.values():
        for classname, lookup in metadata.textures.items():
            lookups[classname] |= lookup

    return lookups



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Self

from pydantic.json_schema import SkipJsonSchema

from hexdoc.core.loader import ModResourceLoader
from hexdoc.core.resource import ResourceLocation
from hexdoc.core.resource_dir import PathResourceDir
from hexdoc.utils import TRACE, JSONDict

from .base import HexdocModel
from .inline import InlineModel

logger = logging.getLogger(__name__)

# TODO: i'm pretty sure there's redundancy between id.py and inline.py


class IDModel(HexdocModel):
    id: SkipJsonSchema[ResourceLocation]
    resource_dir: SkipJsonSchema[PathResourceDir]

    @classmethod
    def load(
        cls,
        resource_dir: PathResourceDir,
        id: ResourceLocation,
        data: JSONDict,
        context: dict[str, Any],
    ) -> Self:
        logger.log(TRACE, f"Load {cls} at {id}")
        return cls.model_validate(
            data | {"id": id, "resource_dir": resource_dir},
            context=context,
        )


class ResourceModel(IDModel, InlineModel, ABC):
    @classmethod
    def load_id(cls, id: ResourceLocation, context: dict[str, Any]):
        loader = ModResourceLoader.of(context)
        resource_dir, data = cls.load_resource(id, loader)
        return cls.load(resource_dir, id, data, context)

    @classmethod
    @abstractmethod
    def load_resource(
        cls,
        id: ResourceLocation,
        loader: ModResourceLoader,
    ) -> tuple[PathResourceDir, JSONDict]: ...



from contextvars import ContextVar
from pathlib import Path
from typing import Annotated

from pydantic import AfterValidator

from .contextmanagers import set_contextvar

_relative_path_root = ContextVar[Path]("_relative_path_root")


def relative_path_root(path: Path):
    return set_contextvar(_relative_path_root, path)


RelativePath = Annotated[
    Path,
    AfterValidator(lambda path: _relative_path_root.get() / path),
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, ClassVar, Iterator, Self

from pydantic import Field

from hexdoc.core import ModResourceLoader, ResourceLocation
from hexdoc.core.resource import BaseResourceLocation
from hexdoc.model import HexdocModel
from hexdoc.utils import PydanticOrderedSet, decode_json_dict


@dataclass
class TagLoader:
    namespace: str
    registry: str
    path: str

    @property
    def id(self):
        return ResourceLocation(self.namespace, self.path)

    def load(self, loader: ModResourceLoader):
        return Tag.load(
            id=self.id,
            registry=self.registry,
            loader=loader,
        )


class OptionalTagValue(HexdocModel, frozen=True):
    id: ResourceLocation
    required: bool


TagValue = ResourceLocation | OptionalTagValue


class Tag(HexdocModel):
    GASLIGHTING_ITEMS: ClassVar = TagLoader("hexdoc", "items", "gaslighting")
    """Item/block ids that gaslight you. This tag isn't real, it's all in your head.

    File: `hexdoc/tags/items/gaslighting.json`
    """
    SPOILERED_ADVANCEMENTS: ClassVar = TagLoader("hexdoc", "advancements", "spoilered")
    """Advancements for entries that should be blurred in the web book.

    File: `hexdoc/tags/advancements/spoilered.json`
    """

    registry: str = Field(exclude=True)
    values: PydanticOrderedSet[TagValue]
    replace: bool = False

    @classmethod
    def load(
        cls,
        registry: str,
        id: ResourceLocation,
        loader: ModResourceLoader,
    ) -> Self:
        values = PydanticOrderedSet[TagValue]()
        replace = False

        for _, _, tag in loader.load_resources(
            "data",
            folder=f"tags/{registry}",
            id=id,
            decode=lambda raw_data: Tag._convert(
                registry=registry,
                raw_data=raw_data,
            ),
            export=cls._export,
        ):
            if tag.replace:
                values.clear()
            for value in tag._load_values(loader):
                values.add(value)

        return cls(registry=registry, values=values, replace=replace)

    @classmethod
    def _convert(cls, *, registry: str, raw_data: str) -> Self:
        data = decode_json_dict(raw_data)
        return cls.model_validate(data | {"registry": registry})

    @property
    def value_ids(self) -> Iterator[ResourceLocation]:
        for value in self.values:
            match value:
                case ResourceLocation():
                    yield value
                case OptionalTagValue(id=id):
                    yield id

    @property
    def value_ids_set(self):
        return set(self.value_ids)

    def __ror__(self, other: set[ResourceLocation]):
        new = set(other)
        new |= self.value_ids_set
        return new

    def __contains__(self, x: Any) -> bool:
        if isinstance(x, BaseResourceLocation):
            return x.id in self.value_ids_set
        return NotImplemented

    def _export(self: Tag, current: Tag | None):
        if self.replace or current is None:
            tag = self
        else:
            tag = self.model_copy(
                update={"values": current.values | self.values},
            )
        return tag.model_dump_json(by_alias=True)

    def _load_values(self, loader: ModResourceLoader) -> Iterator[TagValue]:
        for value in self.values:
            match value:
                case (
                    (ResourceLocation() as child_id)
                    | OptionalTagValue(id=child_id)
                ) if child_id.is_tag:
                    try:
                        child = Tag.load(self.registry, child_id, loader)
                        yield from child._load_values(loader)
                    except FileNotFoundError:
                        yield value
                case _:
                    yield value



import inspect
import logging
import os
import platform
import sys
from importlib.resources import Package
from pathlib import Path
from textwrap import dedent

from hexdoc.__version__ import VERSION
from hexdoc.core import (
    Properties,
)
from hexdoc.plugin import ModPlugin, PluginManager
from hexdoc.plugin.manager import flatten, import_package

logger = logging.getLogger(__name__)


def get_header(
    props: Properties,
    pm: PluginManager,
    mod_plugin: ModPlugin,
):
    python_version = platform.python_version()
    plugins = ", ".join(_plugins(pm))
    jinja_template_roots = ", ".join(_jinja_template_roots(mod_plugin))

    return dedent(
        f"""\
        Python:         {python_version} ({sys.platform})
        hexdoc:         {VERSION}
        plugins:        {plugins}
        mod_plugin:     {mod_plugin.modid}-{mod_plugin.full_version}
        cwd:            {Path.cwd()}
        props_dir:      {_relative_path(props.props_dir)}
        mod_templates:  {jinja_template_roots}
        site_url:       {props.env.github_pages_url}
        """
    ).rstrip()


def _plugins(pm: PluginManager):
    seen = set[str]()
    for _, dist in pm.inner.list_plugin_distinfo():
        if dist.project_name == "hexdoc":
            continue
        project_name = dist.project_name.removeprefix("hexdoc-")
        plugin_name = f"{project_name}-{dist.version}"
        if plugin_name not in seen:
            seen.add(plugin_name)
            yield plugin_name


def _jinja_template_roots(mod_plugin: ModPlugin):
    for package, folder in flatten([mod_plugin.jinja_template_root() or []]):
        module_path = _get_package_path(package)
        folder_path = module_path / folder
        yield _relative_path(folder_path)


def _get_package_path(package: Package):
    module = import_package(package)
    module_path = Path(inspect.getfile(module))
    return module_path.parent if module_path.suffix else module_path


def _relative_path(path: Path):
    cwd = Path.cwd()
    if path == cwd:
        return "."

    try:
        path = path.relative_to(cwd)
    except ValueError:
        return str(path)

    return f".{os.sep}{path}"



import logging
from abc import ABC, abstractmethod
from typing import Iterator

from pydantic import PrivateAttr, ValidationInfo, model_validator

from hexdoc.core import ResourceLocation
from hexdoc.model import HexdocModel

from .book_context import BookContext

logger = logging.getLogger(__name__)


class AdvancementSpoilered(HexdocModel, ABC):
    _is_spoiler: bool = PrivateAttr(False)

    @property
    def is_spoiler(self):
        return self._is_spoiler

    @abstractmethod
    def _get_advancement(self) -> ResourceLocation | None: ...

    @model_validator(mode="after")
    def _check_is_spoiler(self, info: ValidationInfo):
        # consider this object a spoiler if its advancement is in the spoilers tag
        advancement = self._get_advancement()
        if info.context and advancement:
            book_ctx = BookContext.of(info)
            self._is_spoiler = any(
                advancement.match(pattern)
                for pattern in book_ctx.spoilered_advancements
            )

        return self


class Flagged(HexdocModel):
    """Mixin model for categories, entries, and pages to implement flags."""

    flag: str | None = None

    _is_flag_enabled: bool = PrivateAttr(True)

    @property
    def is_flag_enabled(self) -> bool:
        return self._is_flag_enabled

    @model_validator(mode="after")
    def _evaluate_flag(self, info: ValidationInfo):
        ctx = BookContext.of(info)
        self._is_flag_enabled = (
            _evaluate_flag(self.flag, ctx.flags) if self.flag else True
        )
        return self


# https://github.com/VazkiiMods/Patchouli/blob/abd6d03a08c37bcf116730021fda9f477412b31f/Xplat/src/main/java/vazkii/patchouli/common/base/PatchouliConfig.java#L42
def _evaluate_flag(flag: str, flags: dict[str, bool]) -> bool:
    # this SHOULD never be called with an empty string
    match flag[0]:
        case "&":
            return all(_split_and_evaluate_flag(flag, flags))
        case "|":
            return any(_split_and_evaluate_flag(flag, flags))
        case "!":
            flag = flag[1:]
            target = False
        case _:
            target = True

    flag = flag.strip().lower()

    b = flags.get(flag)
    if b is None:
        if flag.startswith("advancements_disabled_"):
            b = False
        else:
            if not flag.startswith("mod:"):
                logger.warning(f"Unknown config flag defaulting to True: {flag}")
            b = True
        # speed up subsequent checks a bit and avoid logging unnecessary warnings
        flags[flag] = b

    return b == target


def _split_and_evaluate_flag(flag: str, flags: dict[str, bool]) -> Iterator[bool]:
    for inner in flag.replace("&", "").replace("|", "").split(","):
        yield _evaluate_flag(inner, flags)



from __future__ import annotations

import json
import logging
import textwrap
from collections import defaultdict
from functools import total_ordering
from typing import Any, Callable, Self

import langcodes
from pydantic import ValidationInfo, model_validator
from pydantic.functional_validators import ModelWrapValidatorHandler

from hexdoc.core import (
    ItemStack,
    ModResourceLoader,
    ResourceLocation,
    ValueIfVersion,
)
from hexdoc.core.properties import LangProps
from hexdoc.model import HexdocModel, ValidationContextModel
from hexdoc.model.base import DEFAULT_CONFIG
from hexdoc.utils import decode_and_flatten_json_dict
from hexdoc.utils.json_schema import inherited, json_schema_extra_config, type_str

logger = logging.getLogger(__name__)


@total_ordering
class LocalizedStr(HexdocModel, frozen=True):
    """Represents a string which has been localized."""

    model_config = DEFAULT_CONFIG | json_schema_extra_config(type_str, inherited)

    key: str
    value: str

    @classmethod
    def skip_i18n(cls, key: str) -> Self:
        """Returns an instance of this class with `value = key`."""
        return cls(key=key, value=key)

    @classmethod
    def with_value(cls, value: str) -> Self:
        """Returns an instance of this class with an empty key."""
        return cls(key="", value=value)

    @model_validator(mode="wrap")
    @classmethod
    def _check_localize(
        cls,
        value: str | Any,
        handler: ModelWrapValidatorHandler[Self],
        info: ValidationInfo,
    ) -> Self:
        # NOTE: if we need LocalizedStr to work as a dict key, add another check which
        # returns cls.skip_i18n(value) if info.context is falsy
        if not isinstance(value, str):
            return handler(value)

        i18n = I18n.of(info)
        return cls._localize(i18n, value)

    @classmethod
    def _localize(cls, i18n: I18n, key: str) -> Self:
        return cls.model_validate(i18n.localize(key))

    def map(self, fn: Callable[[str], str]) -> Self:
        """Returns a copy of this object with `new.value = fn(old.value)`."""
        return self.model_copy(update={"value": fn(self.value)})

    def __repr__(self) -> str:
        return self.value

    def __str__(self) -> str:
        return self.value

    def __eq__(self, other: Self | str | Any):
        match other:
            case LocalizedStr():
                return self.value == other.value
            case str():
                return self.value == other
            case _:
                return super().__eq__(other)

    def __lt__(self, other: Self | str):
        match other:
            case LocalizedStr():
                return self.value < other.value
            case str():
                return self.value < other


class LocalizedItem(LocalizedStr, frozen=True):
    @classmethod
    def _localize(cls, i18n: I18n, key: str) -> Self:
        return cls.model_validate(i18n.localize_item(key))


class I18n(ValidationContextModel):
    """Handles localization of strings."""

    lookup: dict[str, LocalizedStr]
    lang: str
    default_i18n: I18n | None
    enabled: bool
    lang_props: LangProps

    @classmethod
    def list_all(cls, loader: ModResourceLoader):
        # don't list languages which this particular mod doesn't support
        # eg. if Hex has translations for ru_ru but an addon doesn't
        return set(
            id.path
            for resource_dir, id, _ in cls._load_lang_resources(loader)
            if not resource_dir.external
        )

    @classmethod
    def load_all(cls, loader: ModResourceLoader, enabled: bool):
        # lang -> (key -> value)
        lookups = defaultdict[str, dict[str, LocalizedStr]](dict)
        internal_langs = set[str]()

        for resource_dir, lang_id, data in cls._load_lang_resources(loader):
            lang = lang_id.path
            if not langcodes.tag_is_valid(lang):
                modid = resource_dir.modid or lang_id.namespace
                raise ValueError(
                    textwrap.dedent(f"""\
                        Attempted to load invalid lang (provided by {modid}): {lang}
                        Resource dir: {resource_dir.path}
                    """).rstrip()
                )

            lookups[lang] |= cls.parse_lookup(data)
            if not resource_dir.external:
                internal_langs.add(lang)

        default_lang = loader.props.default_lang
        default_lookup = lookups[default_lang]
        default_i18n = cls(
            lookup=default_lookup,
            lang=default_lang,
            default_i18n=None,
            enabled=enabled,
            lang_props=loader.props.lang[default_lang],
        )

        return {default_lang: default_i18n} | {
            lang: cls(
                lookup=lookup,
                lang=lang,
                default_i18n=default_i18n,
                enabled=enabled,
                lang_props=loader.props.lang[lang],
            )
            for lang, lookup in lookups.items()
            if lang in internal_langs and lang != default_lang
        }

    @classmethod
    def load(
        cls,
        loader: ModResourceLoader,
        enabled: bool,
        lang: str,
    ) -> Self:
        if not langcodes.tag_is_valid(lang):
            raise ValueError(f"Invalid lang: {lang}")

        lookup = dict[str, LocalizedStr]()
        is_internal = False

        for resource_dir, _, data in cls._load_lang_resources(loader, lang):
            lookup |= cls.parse_lookup(data)
            if not resource_dir.external:
                is_internal = True

        if enabled and not is_internal:
            raise FileNotFoundError(
                f"Lang {lang} exists, but {loader.props.modid} does not support it"
            )

        default_lang = loader.props.default_lang
        default_i18n = None
        if lang != default_lang:
            default_i18n = cls.load(loader, enabled, default_lang)

        return cls(
            lookup=lookup,
            lang=lang,
            default_i18n=default_i18n,
            enabled=enabled,
            lang_props=loader.props.lang[lang],
        )

    @classmethod
    def parse_lookup(cls, raw_lookup: dict[str, str]) -> dict[str, LocalizedStr]:
        return {
            key: LocalizedStr(key=key, value=value.replace("%%", "%"))
            for key, value in raw_lookup.items()
        }

    @classmethod
    def _load_lang_resources(cls, loader: ModResourceLoader, lang: str = "*"):
        return loader.load_resources(
            "assets",
            namespace="*",
            folder="lang",
            glob=[
                f"{lang}.json",
                f"{lang}.json5",
                f"{lang}.flatten.json",
                f"{lang}.flatten.json5",
            ],
            decode=decode_and_flatten_json_dict,
            export=cls._export,
        )

    @classmethod
    def _export(cls, new: dict[str, str], current: dict[str, str] | None):
        return json.dumps((current or {}) | new)

    @model_validator(mode="after")
    def _warn_if_disabled(self):
        if not self.enabled:
            logger.info(
                f"I18n is disabled for {self.lang}. Warnings about missing translations"
                + " will only be logged in verbose mode."
            )
        elif self.lang_props.quiet:
            logger.info(
                f"Quiet mode is enabled for {self.lang}. Warnings about missing"
                + " translations will only be logged in verbose mode."
            )
        return self

    @property
    def is_default(self):
        return self.default_i18n is None

    def localize(
        self,
        *keys: str,
        default: str | None = None,
        silent: bool = False,
    ) -> LocalizedStr:
        """Looks up the given string in the lang table if i18n is enabled. Otherwise,
        returns the original key.

        If multiple keys are provided, returns the value of the first key which exists.
        That is, subsequent keys are treated as fallbacks for the first.

        Raises ValueError if i18n is enabled and default is None but the key has no
        corresponding localized value.
        """

        for key in keys:
            if key in self.lookup:
                return self.lookup[key]

        if silent or not self.enabled or self.lang_props.quiet:
            log_level = logging.DEBUG
        else:
            log_level = logging.WARNING

        log_keys = set(keys) - self._logged_missing_keys
        if log_keys:
            self._logged_missing_keys.update(log_keys)
            match list(log_keys):
                case [key]:
                    message = f"key {key}"
                case _:
                    message = "keys " + ", ".join(log_keys)
            logger.log(log_level, f"No translation in {self.lang} for {message}")

        if default is not None:
            return LocalizedStr.skip_i18n(default)

        if self.default_i18n:
            return self.default_i18n.localize(*keys, default=default, silent=silent)

        return LocalizedStr.skip_i18n(keys[0])

    def localize_pattern(
        self,
        op_id: ResourceLocation,
        silent: bool = False,
    ) -> LocalizedStr:
        """Localizes the given pattern id (internal name, eg. brainsweep).

        Raises ValueError if i18n is enabled but the key has no localization.
        """
        key_group = ValueIfVersion(">=1.20", "action", "spell")()

        # prefer the book-specific translation if it exists
        return self.localize(
            f"hexcasting.{key_group}.book.{op_id}",
            f"hexcasting.{key_group}.{op_id}",
            silent=silent,
        )

    def localize_item(
        self,
        item: str | ResourceLocation | ItemStack,
        silent: bool = False,
    ) -> LocalizedItem:
        """Localizes the given item resource name.

        Raises ValueError if i18n is enabled but the key has no localization.
        """
        match item:
            case str():
                item = ItemStack.from_str(item)
            case ResourceLocation(namespace=namespace, path=path):
                item = ItemStack(namespace=namespace, path=path)
            case _:
                pass

        localized = self.localize(
            item.i18n_key(),
            item.i18n_key("block"),
            silent=silent,
        )
        return LocalizedItem(key=localized.key, value=localized.value)

    def localize_entity(self, entity: ResourceLocation, type: str | None = None):
        if type:
            entity = type / entity
        return self.localize(entity.i18n_key("entity"))

    def localize_key(self, key: str, silent: bool = False) -> LocalizedStr:
        if not key.startswith("key."):
            key = "key." + key
        return self.localize(key, silent=silent)

    def localize_item_tag(self, tag: ResourceLocation, silent: bool = False):
        localized = self.localize(
            f"tag.{tag.namespace}.{tag.path}",
            f"tag.item.{tag.namespace}.{tag.path}",
            f"tag.block.{tag.namespace}.{tag.path}",
            default=self.fallback_tag_name(tag),
            silent=silent,
        )
        return LocalizedStr(key=localized.key, value=f"Tag: {localized.value}")

    def fallback_tag_name(self, tag: ResourceLocation):
        """Generates a more-or-less reasonable fallback name for a tag.

        For example:
        * `forge:ores` -> Ores
        * `c:saplings/almond` -> Almond Saplings
        * `c:tea_ingredients/gloopy/weak` -> Tea Ingredients, Gloopy, Weak
        """

        if tag.path.count("/") == 1:
            before, after = tag.path.split("/")
            return f"{after} {before}".title()

        return tag.path.replace("_", " ").replace("/", ", ").title()

    def localize_texture(self, texture_id: ResourceLocation, silent: bool = False):
        path = texture_id.path.removeprefix("textures/").removesuffix(".png")
        root, rest = path.split("/", 1)

        # TODO: refactor / extensibilify
        if root == "mob_effect":
            root = "effect"

        return self.localize(f"{root}.{texture_id.namespace}.{rest}", silent=silent)

    def localize_lang(self, silent: bool = False):
        name = self.localize("language.name", default="", silent=silent)
        region = self.localize("language.region", silent=silent)
        # don't allow language names to fall back to English (United States)
        if name.value == "":
            return self.lang
        return f"{name} ({region})"

    @model_validator(mode="after")
    def _init_logger_cache(self):
        self._logged_missing_keys = set[str]()
        return self



__all__ = [
    "JSONDict",
    "JSONValue",
    "TOMLDict",
    "TOMLValue",
    "cast_or_raise",
    "decode_and_flatten_json_dict",
    "decode_json_dict",
    "isinstance_or_raise",
    "load_toml_with_placeholders",
]

from .assertions import cast_or_raise, isinstance_or_raise
from .json import JSONDict, JSONValue, decode_and_flatten_json_dict, decode_json_dict
from .toml import TOMLDict, TOMLValue, load_toml_with_placeholders



from __future__ import annotations

import logging
from collections import defaultdict
from collections.abc import Set
from typing import Annotated, Any, Literal

from hexdoc.core import ModResourceLoader, ResourceLocation
from hexdoc.model import HexdocModel
from hexdoc.utils import JSONDict, clamping_validator

logger = logging.getLogger(__name__)

FoundNormalTexture = tuple[Literal["texture", "block_model"], ResourceLocation]
FoundGaslightingTexture = tuple[Literal["gaslighting"], list[FoundNormalTexture]]
FoundTexture = FoundNormalTexture | FoundGaslightingTexture

ItemDisplayPosition = Literal[
    "thirdperson_righthand",
    "thirdperson_lefthand",
    "firstperson_righthand",
    "firstperson_lefthand",
    "gui",
    "head",
    "ground",
    "fixed",
]

_Translation = Annotated[float, clamping_validator(-80, 80)]
_Scale = Annotated[float, clamping_validator(-4, 4)]


class ItemDisplay(HexdocModel):
    rotation: tuple[float, float, float] | None = None
    translation: tuple[_Translation, _Translation, _Translation] | None = None
    scale: tuple[_Scale, _Scale, _Scale] | None = None


class ModelOverride(HexdocModel):
    model: ResourceLocation
    """The id of the model to use if the case is met."""
    predicate: dict[ResourceLocation, float]


# allow missing because mods can add custom fields :/
class ModelItem(HexdocModel, extra="allow"):
    """https://minecraft.wiki/w/Tutorials/Models#Item_models

    This is called BaseModelItem instead of BaseItemModel because SomethingModel is our
    naming convention for abstract Pydantic models.
    """

    id: ResourceLocation
    """Not in the actual file."""

    parent: ResourceLocation | None = None
    """Loads a different model with the given id."""
    display: dict[ItemDisplayPosition, ItemDisplay] | None = None
    gui_light: Literal["front", "side"] = "side"
    overrides: list[ModelOverride] | None = None
    # TODO: minecraft_render would need to support this
    elements: Any | None = None
    # TODO: support texture variables etc?
    textures: dict[str, ResourceLocation] | None = None
    """Texture ids. For example, `{"layer0": "item/red_bed"}` refers to the resource
    `assets/minecraft/textures/item/red_bed.png`.

    Technically this is only allowed for `minecraft:item/generated`, but we're currently
    not loading Minecraft's item models, so there's lots of other parent ids that this
    field can show up for.
    """

    @classmethod
    def load_resource(cls, id: ResourceLocation, loader: ModResourceLoader):
        _, data = loader.load_resource("assets", "models", id, export=False)
        return cls.load_data(id, data)

    @classmethod
    def load_data(cls, id: ResourceLocation, data: JSONDict):
        return cls.model_validate(data | {"id": id})

    @property
    def item_id(self):
        if "/" not in self.id.path:
            return self.id
        path_without_prefix = "/".join(self.id.path.split("/")[1:])
        return self.id.with_path(path_without_prefix)

    @property
    def layer0(self):
        if self.textures:
            return self.textures.get("layer0")

    def find_texture(
        self,
        loader: ModResourceLoader,
        gaslighting_items: Set[ResourceLocation],
        checked_overrides: defaultdict[ResourceLocation, set[int]] | None = None,
    ) -> FoundTexture | None:
        """May return a texture **or** a model. Texture ids will start with `textures/`."""
        if checked_overrides is None:
            checked_overrides = defaultdict(set)

        # gaslighting
        # as of 0.11.1-7, all gaslighting item models are implemented with overrides
        if self.item_id in gaslighting_items:
            if not self.overrides:
                raise ValueError(
                    f"Model {self.id} for item {self.item_id} marked as gaslighting but"
                    " does not have overrides"
                )

            gaslighting_textures = list[FoundNormalTexture]()
            for i, override in enumerate(self.overrides):
                match self._find_override_texture(
                    i, override, loader, gaslighting_items, checked_overrides
                ):
                    case "gaslighting", _:
                        raise ValueError(
                            f"Model {self.id} for item {self.item_id} marked as"
                            f" gaslighting but override {i} resolves to another gaslighting texture"
                        )
                    case None:
                        break
                    case result:
                        gaslighting_textures.append(result)
            else:
                return "gaslighting", gaslighting_textures

        # if it exists, the layer0 texture is *probably* representative
        # TODO: impl multi-layer textures for Sam
        if self.layer0:
            texture_id = "textures" / self.layer0 + ".png"
            return "texture", texture_id

        # first resolvable override, if any
        for i, override in enumerate(self.overrides or []):
            if result := self._find_override_texture(
                i, override, loader, gaslighting_items, checked_overrides
            ):
                return result

        if self.parent and self.parent.path.startswith("block/"):
            # try the parent id
            # we only do this for blocks in the same namespace because most other
            # parents are generic "base class"-type models which won't actually
            # represent the item
            if self.parent.namespace == self.id.namespace:
                return "block_model", self.parent

            # FIXME: hack
            # this entire selection process needs to be redone, but the idea here is to
            # try rendering item models as blocks in certain cases (eg. edified button)
            return "block_model", self.id

        return None

    def _find_override_texture(
        self,
        index: int,
        override: ModelOverride,
        loader: ModResourceLoader,
        gaslighting_items: Set[ResourceLocation],
        checked_overrides: defaultdict[ResourceLocation, set[int]],
    ) -> FoundTexture | None:
        if override.model.path.startswith("block/"):
            return "block_model", override.model

        if index in checked_overrides[self.id]:
            logger.debug(f"Ignoring recursive override: {override.model}")
            return None

        checked_overrides[self.id].add(index)
        return (
            (ModelItem)
            .load_resource(override.model, loader)
            .find_texture(loader, gaslighting_items, checked_overrides)
        )



from hexdoc.core import ResourceLocation
from hexdoc.minecraft import LocalizedStr
from hexdoc.model import HexdocModel
from pydantic import Field, model_validator


def _gui_texture(name: str):
    return ResourceLocation("modonomicon", f"textures/gui/{name}.png")


class Modonomicon(HexdocModel):
    """https://klikli-dev.github.io/modonomicon/docs/basics/structure/book"""

    name: LocalizedStr

    tooltip: LocalizedStr | None = None
    generate_book_item: bool = True
    model: ResourceLocation | None = Field("modonomicon:modonomicon_purple")  # type: ignore
    custom_book_item: ResourceLocation | None = None
    creative_tab: str = "misc"
    default_title_color: int = 0
    auto_add_read_conditions: bool = False
    book_overview_texture: ResourceLocation = _gui_texture("book_overview")
    frame_texture: ResourceLocation = _gui_texture("book_frame")
    left_frame_overlay: ResourceLocation = _gui_texture("book_frame_left_overlay")
    right_frame_overlay: ResourceLocation = _gui_texture("book_frame_right_overlay")
    top_frame_overlay: ResourceLocation = _gui_texture("book_frame_top_overlay")
    bottom_frame_overlay: ResourceLocation = _gui_texture("book_frame_bottom_overlay")
    book_content_texture: ResourceLocation = _gui_texture("book_content")
    crafting_texture: ResourceLocation = _gui_texture("crafting_textures")
    category_button_icon_scale: float = 1
    category_button_x_offset: int = 0
    category_button_y_offset: int = 0
    search_button_x_offset: int = 0
    search_button_y_offset: int = 0
    read_all_button_y_offset: int = 0
    turn_page_sound: ResourceLocation = Field("minecraft:turn_page")  # type: ignore

    @model_validator(mode="after")
    def _validate_constraints(self):
        if self.generate_book_item:
            assert self.model, "model is required if generate_book_item is True"
        else:
            assert (
                self.custom_book_item
            ), "custom_book_item is required if generate_book_item is False"

        return self



from __future__ import annotations

from typing import Any, Callable, Generic, TypeVar

_T_cv = TypeVar("_T_cv", contravariant=True)
_R_co = TypeVar("_R_co", covariant=True)


# https://discuss.python.org/t/add-a-supported-read-only-classproperty-decorator-in-the-stdlib/18090
class ClassPropertyDescriptor(Generic[_T_cv, _R_co]):
    """Equivalent of `classmethod(property(...))`.

    Use `@classproperty`. Do not instantiate this class directly.
    """

    def __init__(self, func: classmethod[_T_cv, Any, _R_co]) -> None:
        self.func = func

    def __get__(self, _: Any, cls: type[_T_cv]) -> _R_co:
        return self.func.__func__(cls)


def classproperty(
    func: Callable[[type[_T_cv]], _R_co],
) -> ClassPropertyDescriptor[_T_cv, _R_co]:
    if isinstance(func, classmethod):
        return ClassPropertyDescriptor(func)
    return ClassPropertyDescriptor(classmethod(func))



from typing import Any, ClassVar, Iterator, Unpack

from pydantic import ConfigDict, Field, PrivateAttr, ValidationInfo, model_validator
from typing_extensions import override

from hexdoc.core import (
    ModResourceLoader,
    ResourceLocation,
    ValueIfVersion,
)
from hexdoc.core.compat import AtLeast_1_20, Before_1_20
from hexdoc.model import ResourceModel, TypeTaggedTemplate
from hexdoc.utils import Inherit, InheritType, classproperty

from ..assets import ImageTexture, ItemWithTexture, validate_texture
from ..i18n import I18n, LocalizedStr
from .ingredients import ItemIngredient, ItemIngredientList, ItemResult


class Recipe(TypeTaggedTemplate, ResourceModel):
    """Base model for Minecraft recipes.

    https://minecraft.wiki/w/Recipe
    """

    category: str | None = None
    group: str | None = None
    show_notification: AtLeast_1_20[bool] | Before_1_20[None] = Field(
        default_factory=ValueIfVersion(">=1.20", True, None)
    )

    fabric_load_conditions: list[dict[str, Any]] | None = Field(
        default=None, alias="fabric:load_conditions"
    )
    """Fabric resource conditions.

    https://github.com/FabricMC/fabric/blob/761f669d0a6fbfe2ae6d71d767651f32a13d37fc/fabric-resource-conditions-api-v1/src/main/java/net/fabricmc/fabric/api/resource/conditions/v1/ResourceConditions.java#L64
    """

    # not in the actual model

    _workstation: ClassVar[ResourceLocation | None] = None

    _gui_name: LocalizedStr | None = PrivateAttr(None)
    _gui_texture: ImageTexture | None = PrivateAttr(None)

    def __init_subclass__(
        cls,
        *,
        type: str | InheritType | None = Inherit,
        workstation: str | InheritType | None = Inherit,
        **kwargs: Unpack[ConfigDict],
    ):
        super().__init_subclass__(type=type, **kwargs)

        match workstation:
            case str():
                cls._workstation = ResourceLocation.from_str(workstation)
            case None:
                cls._workstation = None
            case _:
                pass

    @classmethod
    def load_resource(cls, id: ResourceLocation, loader: ModResourceLoader):
        return loader.load_resource("data", "recipes", id)

    @classproperty
    @classmethod
    @override
    def template(cls):
        return cls.template_id.template_path("recipes")

    @classproperty
    @classmethod
    @override
    def template_id(cls):
        assert cls._workstation is not None
        return cls._workstation

    @property
    def gui_name(self):
        return self._gui_name

    @property
    def gui_texture(self):
        return self._gui_texture

    @classproperty
    @classmethod
    def _gui_texture_id(cls) -> ResourceLocation | None:
        """ResourceLocation of the background image for this recipe type."""
        if cls._workstation is None:
            return None
        return ResourceLocation(
            cls._workstation.namespace,
            f"textures/gui/hexdoc/{cls._workstation.path}.png",
        )

    def _localize_workstation(self, i18n: I18n):
        if self._workstation is not None:
            return i18n.localize_item(self._workstation)

    @model_validator(mode="after")
    def _load_gui_texture(self, info: ValidationInfo):
        self._gui_name = self._localize_workstation(I18n.of(info))

        if self._gui_texture_id is not None:
            self._gui_texture = validate_texture(
                self._gui_texture_id,
                context=info,
                model_type=ImageTexture,
            )

        return self


class CraftingRecipe(Recipe, workstation="minecraft:crafting_table"):
    result: ItemResult


class CraftingShapelessRecipe(CraftingRecipe, type="minecraft:crafting_shapeless"):
    ingredients: list[ItemIngredientList]


class CraftingShapedRecipe(CraftingRecipe, type="minecraft:crafting_shaped"):
    key: dict[str, ItemIngredientList]
    pattern: list[str]

    @property
    def ingredients(self) -> Iterator[ItemIngredientList | None]:
        for row in self.pattern:
            if len(row) > 3:
                raise ValueError(f"Expected len(row) <= 3, got {len(row)}: `{row}`")
            for item_key in row.ljust(3):
                match item_key:
                    case " ":
                        yield None
                    case _:
                        yield self.key[item_key]


class CookingRecipe(Recipe):
    ingredient: ItemIngredientList
    result: ItemWithTexture
    experience: float
    cookingtime: int


class BlastingRecipe(
    CookingRecipe,
    type="minecraft:blasting",
    workstation="minecraft:blast_furnace",
):
    cookingtime: int = 100


class CampfireCookingRecipe(
    CookingRecipe,
    type="minecraft:campfire_cooking",
    workstation="minecraft:campfire",
):
    cookingtime: int = 100


class SmeltingRecipe(
    CookingRecipe,
    type="minecraft:smelting",
    workstation="minecraft:furnace",
):
    cookingtime: int = 200


class SmokingRecipe(
    CookingRecipe,
    type="minecraft:smoking",
    workstation="minecraft:smoker",
):
    cookingtime: int = 100


class SmithingRecipe(Recipe, workstation="minecraft:smithing_table"):
    base: ItemIngredient
    addition: ItemIngredient
    template_ingredient: ItemIngredient = Field(alias="template")

    @property
    def result_item(self):
        return self.base.item


class SmithingTransformRecipe(SmithingRecipe, type="minecraft:smithing_transform"):
    result: ItemWithTexture

    @property
    def result_item(self):
        return self.result


class SmithingTrimRecipe(SmithingRecipe, type="minecraft:smithing_trim"):
    pass


class StonecuttingRecipe(
    Recipe,
    type="minecraft:stonecutting",
    workstation="minecraft:stonecutter",
):
    ingredient: ItemIngredientList
    result: ItemWithTexture
    count: int



import logging
import os
import re
import sys
from bisect import bisect
from logging import (
    Filter,
    Formatter,
    LogRecord,
    StreamHandler,
)
from typing import Any, Iterable, Literal, Mapping

from .tracebacks import create_filtered_excepthook

TRACE = 5
"""For even more verbose logs than `logging.DEBUG`."""

TRACEBACK_HIDDEN_MODULES = {
    "click.core",
    "typer.core",
    "typer.main",
}


logger = logging.getLogger(__name__)

filtered_excepthook = create_filtered_excepthook(TRACEBACK_HIDDEN_MODULES)


class RegexFilter(Filter):
    def __init__(self, pattern: str | re.Pattern[str]):
        super().__init__()
        self.regex: re.Pattern[str]
        match pattern:
            case str():
                self.regex = re.compile(pattern)
            case re.Pattern():
                self.regex = pattern

    def filter(self, record: LogRecord) -> bool:
        if not super().filter(record):
            return False
        return self.regex.search(record.msg) is None


# https://stackoverflow.com/a/68154386
class LevelFormatter(Formatter):
    def __init__(
        self,
        formats: dict[int, str],
        datefmt: str | None = None,
        style: Literal["%", "{", "$"] = "%",
        validate: bool = True,
        *,
        defaults: Mapping[str, Any] | None = None,
    ):
        super().__init__()

        self.formats = sorted(
            (
                level,
                Formatter(fmt, datefmt, style, validate, defaults=defaults),
            )
            for level, fmt in formats.items()
        )

    def format(self, record: LogRecord) -> str:
        idx = bisect(self.formats, (record.levelno,), hi=len(self.formats) - 1)
        _, formatter = self.formats[idx]
        return formatter.format(record)


# separate class so we can isinstance below
class HexdocLevelFormatter(LevelFormatter):
    pass


def setup_logging(
    verbosity: int,
    ci: bool,
    *,
    filters: Iterable[Filter] | None = None,
    quiet_langs: Iterable[str] | None = None,
):
    logging.addLevelName(TRACE, "TRACE")

    root_logger = logging.getLogger()

    if root_logger.handlers:
        for handler in root_logger.handlers:
            if isinstance(handler.formatter, HexdocLevelFormatter):
                logger.debug(f"Removing existing handler from root logger: {handler}")
                root_logger.removeHandler(handler)

    level = verbosity_log_level(verbosity)
    root_logger.setLevel(level)

    formats = {
        logging.DEBUG: log_format("relativeCreated", "levelname", "name"),
    }

    if level >= logging.INFO:
        # set this here so we don't clobber exceptions in verbose mode
        # but don't set it if Typer's pretty tracebacks are enabled
        # see also: https://typer.tiangolo.com/tutorial/exceptions/#disable-pretty-exceptions
        if os.getenv("_TYPER_STANDARD_TRACEBACK"):
            sys.excepthook = filtered_excepthook

        formats |= {
            logging.INFO: log_format("levelname"),
            logging.WARNING: log_format("levelname", "name"),
        }

    if ci:
        formats |= {
            logging.WARNING: "::warning file={name},line={lineno},title={levelname}::{message}",
            logging.ERROR: "::error file={name},line={lineno},title={levelname}::{message}",
        }

    handler = StreamHandler()

    handler.setLevel(level)
    handler.setFormatter(HexdocLevelFormatter(formats, style="{"))

    if filters:
        for filter in filters:
            handler.addFilter(filter)

    if quiet_langs:
        for lang in quiet_langs:
            handler.addFilter(RegexFilter(f"^No translation in {lang}"))

    root_logger.addHandler(handler)

    logging.getLogger("PIL").setLevel(logging.INFO)

    logger.debug("Initialized logger.")


def log_format(*names: Literal["relativeCreated", "levelname", "name"]):
    components = {
        "relativeCreated": "{relativeCreated:.02f}",
        "levelname": "{levelname}",
        "name": "{name}",
    }

    joined_components = " | ".join(components[name] for name in names)
    return "\033[1m[" + joined_components + "]\033[0m {message}"


def verbosity_log_level(verbosity: int) -> int:
    match verbosity:
        case 0:
            return logging.INFO
        case 1:
            return logging.DEBUG
        case _:
            return TRACE


def repl_readfunc():
    exit_next = False

    def inner(prompt: str) -> str:
        nonlocal exit_next
        try:
            response = input(prompt)
            exit_next = False
            return response
        except KeyboardInterrupt:
            if exit_next:
                print("\nExiting.")
                sys.exit()

            print("\nPress ctrl+c again to exit.", end="")
            exit_next = True
            raise

    return inner



__all__ = [
    "DEFAULT_CONFIG",
    "Color",
    "HexdocModel",
    "HexdocSettings",
    "HexdocTypeAdapter",
    "IDModel",
    "InlineItemModel",
    "InlineModel",
    "InternallyTaggedUnion",
    "NoValue",
    "NoValueType",
    "ResourceModel",
    "StripHiddenModel",
    "TagValue",
    "TypeTaggedTemplate",
    "TypeTaggedUnion",
    "ValidationContextModel",
    "init_context",
]

from .base import (
    DEFAULT_CONFIG,
    HexdocModel,
    HexdocSettings,
    HexdocTypeAdapter,
    ValidationContextModel,
    init_context,
)
from .id import IDModel, ResourceModel
from .inline import InlineItemModel, InlineModel
from .strip_hidden import StripHiddenModel
from .tagged_union import (
    InternallyTaggedUnion,
    NoValue,
    NoValueType,
    TagValue,
    TypeTaggedTemplate,
    TypeTaggedUnion,
)
from .types import Color



"""Static values for one way import."""

SUPPORTED_FORMATS = (".svg", ".jpeg", ".jpg", ".png", ".tiff", ".tif")

HTML_LINK = '<link rel="{rel}" type="{type}" href="{href}" />'

ICON_TYPES = (
    {"image_fmt": "ico", "rel": None, "dimensions": (64, 64), "prefix": "favicon"},
    {"image_fmt": "png", "rel": "icon", "dimensions": (16, 16), "prefix": "favicon"},
    {"image_fmt": "png", "rel": "icon", "dimensions": (32, 32), "prefix": "favicon"},
    {"image_fmt": "png", "rel": "icon", "dimensions": (64, 64), "prefix": "favicon"},
    {"image_fmt": "png", "rel": "icon", "dimensions": (96, 96), "prefix": "favicon"},
    {"image_fmt": "png", "rel": "icon", "dimensions": (180, 180), "prefix": "favicon"},
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (57, 57),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (60, 60),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (72, 72),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (76, 76),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (114, 114),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (120, 120),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (144, 144),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (152, 152),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (167, 167),
        "prefix": "apple-touch-icon",
    },
    {
        "image_fmt": "png",
        "rel": "apple-touch-icon",
        "dimensions": (180, 180),
        "prefix": "apple-touch-icon",
    },
    {"image_fmt": "png", "rel": None, "dimensions": (70, 70), "prefix": "mstile"},
    {"image_fmt": "png", "rel": None, "dimensions": (270, 270), "prefix": "mstile"},
    {"image_fmt": "png", "rel": None, "dimensions": (310, 310), "prefix": "mstile"},
    {"image_fmt": "png", "rel": None, "dimensions": (310, 150), "prefix": "mstile"},
    {"image_fmt": "png", "rel": "shortcut icon", "dimensions": (196, 196), "prefix": "favicon"},
)



import shutil
import textwrap
from collections import defaultdict
from pathlib import Path

from packaging.version import Version
from pydantic import Field, TypeAdapter
from pydantic.alias_generators import to_camel

from hexdoc.model import HexdocModel
from hexdoc.utils import write_to_path

MARKER_NAME = ".sitemap-marker.json"


class BaseSitemapMarker(HexdocModel):
    version: str
    lang: str
    lang_name: str
    path: str
    is_default_lang: bool
    full_version: str
    minecraft_version: str | None
    redirect_contents: str
    """Pre-rendered HTML file to create a redirect for this page."""

    @classmethod
    def load(cls, path: Path):
        return cls.model_validate_json(path.read_text("utf-8"))


class VersionedSitemapMarker(BaseSitemapMarker):
    mod_version: str
    plugin_version: str

    def __gt__(self, other: BaseSitemapMarker) -> bool:
        match other:
            case VersionedSitemapMarker():
                return Version(self.plugin_version) > Version(other.plugin_version)
            case LatestSitemapMarker():
                return True
            case _:
                return NotImplemented


class LatestSitemapMarker(BaseSitemapMarker):
    branch: str
    is_default_branch: bool

    def __gt__(self, other: BaseSitemapMarker) -> bool:
        match other:
            case VersionedSitemapMarker():
                return False
            case LatestSitemapMarker():
                return self.is_default_branch and not other.is_default_branch
            case _:
                return NotImplemented


SitemapMarker = VersionedSitemapMarker | LatestSitemapMarker


# TODO: there should be a VersionedSitemapItem and a LatestSitemapItem
class SitemapItem(HexdocModel, alias_generator=to_camel):
    default_lang: str = ""
    default_path: str = ""

    markers: dict[str, SitemapMarker] = Field(default_factory=dict)
    lang_names: dict[str, str] = Field(default_factory=dict)
    lang_paths: dict[str, str] = Field(default_factory=dict)

    @property
    def default_marker(self):
        return self.markers[self.default_lang]

    def add_marker(self, marker: SitemapMarker):
        if (old_marker := self.markers.get(marker.lang)) and old_marker > marker:
            return

        self.markers[marker.lang] = marker
        self.lang_paths[marker.lang] = marker.path
        self.lang_names[marker.lang] = marker.lang_name or marker.lang

        if marker.is_default_lang:
            self.default_lang = marker.lang
            self.default_path = marker.path


Sitemap = dict[str, SitemapItem]
MinecraftSitemap = dict[str, Sitemap]


def delete_updated_books(*, src: Path, dst: Path, release: bool):
    src_markers = src.rglob(MARKER_NAME)
    for marker in src_markers:
        # eg. /v/latest/main/en_us/.sitemap-marker.json -> /v/latest/main
        src_dir = marker.parent.parent
        dst_dir = dst / src_dir.relative_to(src)
        if dst_dir.exists():
            if release:
                raise ValueError(
                    textwrap.dedent(
                        f"""
                        Release mode is enabled, refusing to overwrite existing directory: {dst_dir}
                            Note: this combination of mod and book version has already been released.
                            Try incrementing the book version number (usually in __version__.py).
                        """
                    ).strip()
                )
            shutil.rmtree(dst_dir)


def load_sitemap(root: Path):
    sitemap: Sitemap = defaultdict(SitemapItem)
    minecraft_sitemap: MinecraftSitemap = defaultdict(lambda: defaultdict(SitemapItem))

    # crawl the new tree to rebuild the sitemap
    for marker_path in root.rglob(MARKER_NAME):
        try:
            marker = VersionedSitemapMarker.load(marker_path)
        except ValueError:
            marker = LatestSitemapMarker.load(marker_path)
        sitemap[marker.version].add_marker(marker)
        # TODO: ew.
        minecraft_sitemap[marker.minecraft_version or "???"][marker.version].add_marker(
            marker
        )

    return sitemap, minecraft_sitemap


def dump_sitemap(root: Path, sitemap: Sitemap, minecraft_sitemap: MinecraftSitemap):
    # dump the sitemaps using a TypeAdapter so they serialize the items properly
    write_to_path(
        root / "meta" / "sitemap.json",
        TypeAdapter(Sitemap).dump_json(sitemap, by_alias=True),
    )

    write_to_path(
        root / "meta" / "sitemap-minecraft.json",
        TypeAdapter(MinecraftSitemap).dump_json(minecraft_sitemap, by_alias=True),
    )



from hexdoc.core import ModResourceLoader, ResourceLocation

from .block import BlockModel
from .item import ItemModel


def load_model(loader: ModResourceLoader, model_id: ResourceLocation):
    match model_id.path.split("/")[0]:
        case "block":
            model_type = BlockModel
        case "item":
            model_type = ItemModel
        case type_name:
            raise ValueError(f"Unsupported type {type_name} for model {model_id}")

    try:
        return loader.load_resource(
            type="assets",
            folder="models",
            id=model_id,
            decode=model_type.model_validate_json,
        )
    except Exception as e:
        e.add_note(f"  note: {model_id=}")
        raise



__all__ = [
    "I18n",
    "LocalizedItem",
    "LocalizedStr",
    "Tag",
    "TagValue",
    "assets",
    "recipe",
]

from . import assets, recipe
from .i18n import I18n, LocalizedItem, LocalizedStr
from .tags import Tag, TagValue



import code
import logging
import os
import shutil
import sys
from dataclasses import dataclass
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path
from textwrap import dedent
from typing import Annotated, Any, Optional

import typer
from packaging.version import Version
from typer import Argument, Option
from yarl import URL

from hexdoc.__version__ import VERSION
from hexdoc.core import ModResourceLoader, ResourceLocation
from hexdoc.data.metadata import HexdocMetadata
from hexdoc.data.sitemap import (
    SitemapMarker,
    delete_updated_books,
    dump_sitemap,
    load_sitemap,
)
from hexdoc.graphics.render import BlockRenderer, DebugType
from hexdoc.jinja.render import create_jinja_env, get_templates, render_book
from hexdoc.minecraft import I18n
from hexdoc.minecraft.assets import (
    AnimatedTexture,
    PNGTexture,
    TextureContext,
)
from hexdoc.minecraft.assets.load_assets import render_block
from hexdoc.minecraft.models.item import ItemModel
from hexdoc.minecraft.models.load import load_model
from hexdoc.patchouli import BookContext, FormattingContext
from hexdoc.plugin import ModPluginWithBook
from hexdoc.utils import git_root, setup_logging, write_to_path
from hexdoc.utils.logging import repl_readfunc

from . import ci
from .utils.args import (
    DEFAULT_MERGE_DST,
    DEFAULT_MERGE_SRC,
    BranchOption,
    DefaultTyper,
    PathArgument,
    PropsOption,
    ReleaseOption,
    VerbosityOption,
)
from .utils.load import (
    init_context,
    load_common_data,
    render_textures_and_export_metadata,
)

logger = logging.getLogger(__name__)


def set_default_env():
    """Sets placeholder values for unneeded environment variables."""
    for key, value in {
        "GITHUB_REPOSITORY": "placeholder/placeholder",
        "GITHUB_SHA": "",
        "GITHUB_PAGES_URL": "",
    }.items():
        os.environ.setdefault(key, value)


@dataclass(kw_only=True)
class LoadedBookInfo:
    language: str
    i18n: I18n
    context: dict[str, Any]
    book_id: ResourceLocation
    book: Any


app = DefaultTyper()
app.add_typer(ci.app)


def version_callback(value: bool):
    if value:
        print(f"hexdoc {VERSION}")
        raise typer.Exit()


@app.callback()
def callback(
    verbosity: VerbosityOption = 0,
    quiet_lang: Optional[list[str]] = None,
    version: Annotated[
        bool,
        Option("--version", "-V", callback=version_callback, is_eager=True),
    ] = False,
):
    if quiet_lang:
        logger.warning(
            "`--quiet-lang` is deprecated, use `props.lang.{lang}.quiet` instead."
        )
    setup_logging(verbosity, ci=False, quiet_langs=quiet_lang)


@app.command()
def repl(*, props_file: PropsOption):
    """Start a Python shell with some helpful extra locals added from hexdoc."""

    repl_locals = dict[str, Any](
        props_path=props_file,
    )

    try:
        props, pm, book_plugin, plugin = load_common_data(props_file, branch="")
        repl_locals |= dict(
            props=props,
            pm=pm,
            plugin=plugin,
        )

        loader = ModResourceLoader.load_all(
            props,
            pm,
            export=False,
        )
        repl_locals["loader"] = loader

        if props.book_id:
            book_id, book_data = book_plugin.load_book_data(props.book_id, loader)
        else:
            book_id = None
            book_data = {}

        i18n = I18n.load_all(
            loader,
            enabled=book_plugin.is_i18n_enabled(book_data),
        )[props.default_lang]

        all_metadata = loader.load_metadata(model_type=HexdocMetadata)
        repl_locals["all_metadata"] = all_metadata

        if book_id and book_data:
            context = init_context(
                book_id=book_id,
                book_data=book_data,
                pm=pm,
                loader=loader,
                i18n=i18n,
                all_metadata=all_metadata,
            )
            book = book_plugin.validate_book(book_data, context=context)
            repl_locals |= dict(
                book=book,
                context=context,
            )
    except Exception as e:
        print(e)

    code.interact(
        banner=dedent(
            f"""\
            [hexdoc repl] Python {sys.version}
            Locals: {', '.join(sorted(repl_locals.keys()))}"""
        ),
        readfunc=repl_readfunc(),
        local=repl_locals,
        exitmsg="",
    )


@app.command()
def build(
    output_dir: PathArgument = DEFAULT_MERGE_SRC,
    *,
    branch: BranchOption,
    release: ReleaseOption = False,
    clean: bool = False,
    props_file: PropsOption,
) -> Path:
    """Export resources and render the web book.

    For developers: returns the site path (eg. `/v/latest/main`).
    """

    props, pm, book_plugin, plugin = load_common_data(props_file, branch)

    if props.env.hexdoc_subdirectory:
        output_dir /= props.env.hexdoc_subdirectory

    logger.info("Exporting resources.")
    with ModResourceLoader.clean_and_load_all(props, pm, export=True) as loader:
        site_path = plugin.site_path(versioned=release)
        site_dir = output_dir / site_path

        asset_loader = plugin.asset_loader(
            loader=loader,
            site_url=props.env.github_pages_url.joinpath(*site_path.parts),
            asset_url=props.env.asset_url,
            render_dir=site_dir,
        )

        all_metadata = render_textures_and_export_metadata(loader, asset_loader)

        if not props.book_id:
            logger.info("Skipping book load because props.book_id is not set.")
            return site_dir

        book_id, book_data = book_plugin.load_book_data(props.book_id, loader)

        all_i18n = I18n.load_all(
            loader,
            enabled=book_plugin.is_i18n_enabled(book_data),
        )

        logger.info("Loading books for all languages.")
        books = list[LoadedBookInfo]()
        for language, i18n in all_i18n.items():
            try:
                context = init_context(
                    book_id=book_id,
                    book_data=book_data,
                    pm=pm,
                    loader=loader,
                    i18n=i18n,
                    all_metadata=all_metadata,
                )
                book = book_plugin.validate_book(book_data, context=context)
                books.append(
                    LoadedBookInfo(
                        language=language,
                        i18n=i18n,
                        context=context,
                        book_id=book_id,
                        book=book,
                    )
                )
            except Exception:
                if not props.lang[language].ignore_errors:
                    raise
                logger.exception(f"Failed to load book for {language}")

        if not props.template:
            logger.info("Skipping book render because props.template is not set.")
            return site_dir

        if not isinstance(plugin, ModPluginWithBook):
            raise ValueError(
                f"ModPlugin registered for modid `{props.modid}` (from props.modid)"
                f" does not inherit from ModPluginWithBook: {plugin}"
            )

        logger.info("Setting up Jinja template environment.")
        env = create_jinja_env(pm, props.template.include, props_file)

        logger.info(f"Rendering book for {len(books)} language(s).")

        site_dir.mkdir(parents=True, exist_ok=True)
        pm.pre_render_site(books, env, site_dir, props.template.include)

        for book_info in books:
            try:
                templates = get_templates(
                    props=props,
                    pm=pm,
                    book=book_info.book,
                    context=book_info.context,
                    env=env,
                )
                if not templates:
                    raise RuntimeError(
                        "No templates to render, check your props.template configuration "
                        f"(in {props_file.as_posix()})"
                    )

                book_ctx = BookContext.of(book_info.context)
                formatting_ctx = FormattingContext.of(book_info.context)
                texture_ctx = TextureContext.of(book_info.context)

                site_book_path = plugin.site_book_path(
                    book_info.language,
                    versioned=release,
                )
                if clean:
                    shutil.rmtree(output_dir / site_book_path, ignore_errors=True)

                template_args: dict[str, Any] = book_info.context | {
                    "all_metadata": all_metadata,
                    "png_textures": PNGTexture.get_lookup(texture_ctx.textures),
                    "animations": sorted(  # this MUST be sorted to avoid flaky tests
                        AnimatedTexture.get_lookup(texture_ctx.textures).values(),
                        key=lambda t: t.css_class,
                    ),
                    "book": book_info.book,
                    "book_links": book_ctx.book_links,
                }

                render_book(
                    props=props,
                    pm=pm,
                    plugin=plugin,
                    lang=book_info.language,
                    book_id=book_info.book_id,
                    i18n=book_info.i18n,
                    macros=formatting_ctx.macros,
                    env=env,
                    templates=templates,
                    output_dir=output_dir,
                    version=plugin.mod_version if release else f"latest/{branch}",
                    site_path=site_book_path,
                    versioned=release,
                    template_args=template_args,
                )
            except Exception:
                if not props.lang[book_info.language].ignore_errors:
                    raise
                logger.exception(f"Failed to render book for {book_info.language}")

        pm.post_render_site(books, env, site_dir, props.template.include)

    logger.info("Done.")
    return site_dir


@app.command()
def merge(
    *,
    props_file: PropsOption,
    src: Path = DEFAULT_MERGE_SRC,
    dst: Path = DEFAULT_MERGE_DST,
    release: ReleaseOption = False,
):
    props, _, _, plugin = load_common_data(props_file, branch="", book=True)
    if not props.template:
        raise ValueError("Expected a value for props.template, got None")

    if props.env.hexdoc_subdirectory:
        src /= props.env.hexdoc_subdirectory
        dst /= props.env.hexdoc_subdirectory

    dst.mkdir(parents=True, exist_ok=True)

    # remove any stale data that we're about to replace
    delete_updated_books(src=src, dst=dst, release=release)

    # do the merge
    shutil.copytree(src=src, dst=dst, dirs_exist_ok=True)

    # rebuild the sitemap
    sitemap, minecraft_sitemap = load_sitemap(dst)
    dump_sitemap(dst, sitemap, minecraft_sitemap)

    # find paths for redirect pages
    redirects = dict[Path, str]()

    root_version: Version | None = None
    root_redirect: str | None = None

    def add_redirect(marker: SitemapMarker, path: Path):
        if path != Path(marker.path.removeprefix("/")):
            redirects[path] = marker.redirect_contents

    for version, item in sitemap.items():
        if version.startswith("latest"):  # TODO: check type of item instead
            continue

        add_redirect(item.default_marker, plugin.site_root / version)
        for lang, marker in item.markers.items():
            add_redirect(marker, plugin.site_root / version / lang)

        item_version = Version(version)
        if not root_version or item_version > root_version:
            root_version = item_version
            root_redirect = item.default_marker.redirect_contents

    if root_redirect is None:
        # TODO: use plugin to build this path
        # TODO: refactor
        if item := sitemap.get(f"latest/{props.default_branch}"):
            root_redirect = item.default_marker.redirect_contents
        elif sitemap:
            key = sorted(sitemap.keys())[0]
            root_redirect = sitemap[key].default_marker.redirect_contents
            logger.warning(
                f"No book exists for the default branch `{props.default_branch}`, generating root redirect to `{key}` (check the value of `default_branch` in hexdoc.toml)"
            )
        else:
            logger.error("No books found, skipping root redirect")

    if root_redirect is not None:
        redirects[Path()] = root_redirect

    # write redirect pages
    if props.template.redirect:
        filename, _ = props.template.redirect
        for path, redirect_contents in redirects.items():
            write_to_path(dst / path / filename, redirect_contents)

    # bypass Jekyll on GitHub Pages
    (dst / ".nojekyll").touch()


@app.command()
def serve(
    *,
    props_file: PropsOption,
    port: int = 8000,
    src: Path = DEFAULT_MERGE_SRC,
    dst: Path = DEFAULT_MERGE_DST,
    branch: BranchOption,
    release: bool = True,
    clean: bool = False,
    do_merge: Annotated[bool, Option("--merge/--no-merge")] = True,
):
    book_root = dst
    relative_root = book_root.resolve().relative_to(Path.cwd())

    base_url = URL.build(scheme="http", host="localhost", port=port)
    book_url = base_url.joinpath(*relative_root.parts)

    repo_root = git_root(props_file.parent)
    asset_root = repo_root.relative_to(Path.cwd())

    os.environ |= {
        # prepend a slash to the path so it can find the texture in the local repo
        # eg. http://localhost:8000/_site/src/docs/Common/...
        # vs. http://localhost:8000/Common/...
        "DEBUG_GITHUBUSERCONTENT": str(base_url.joinpath(*asset_root.parts)),
        "GITHUB_PAGES_URL": str(book_url),
    }

    print()
    logger.info(f"hexdoc build --{'' if release else 'no-'}release")
    build(
        branch=branch,
        props_file=props_file,
        output_dir=src,
        release=True,
        clean=clean,
    )

    if do_merge:
        print()
        logger.info("hexdoc merge")
        merge(
            src=src,
            dst=dst,
            props_file=props_file,
        )

    print()
    logger.info(f"Serving web book at {book_url} (press ctrl+c to exit)\n")
    with HTTPServer(("", port), SimpleHTTPRequestHandler) as httpd:
        # ignore KeyboardInterrupt to stop Typer from printing "Aborted."
        # because it keeps printing after nodemon exits and breaking the output
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            pass


@app.command()
def render_model(
    model_id: str,
    *,
    props_file: PropsOption,
    output_path: Annotated[Path, Option("--output", "-o")] = Path("out.png"),
    axes: bool = False,
    normals: bool = False,
):
    """Use hexdoc's block rendering to render an item or block model."""
    set_default_env()
    props, pm, *_ = load_common_data(props_file, branch="")

    debug = DebugType.NONE
    if axes:
        debug |= DebugType.AXES
    if normals:
        debug |= DebugType.NORMALS

    with ModResourceLoader.load_all(props, pm, export=False) as loader:
        _, model = load_model(loader, ResourceLocation.from_str(model_id))
        while isinstance(model, ItemModel) and model.parent:
            _, model = load_model(loader, model.parent)

        if isinstance(model, ItemModel):
            raise ValueError(f"Invalid block id: {model_id}")

        with BlockRenderer(loader=loader, debug=debug) as renderer:
            renderer.render_block_model(model, output_path)


@app.command()
def render_models(
    model_ids: Annotated[Optional[list[str]], Argument()] = None,
    *,
    props_file: PropsOption,
    render_all: Annotated[bool, Option("--all")] = False,
    output_dir: Annotated[Path, Option("--output", "-o")] = Path("out"),
    site_url_str: Annotated[Optional[str], Option("--site-url")] = None,
    export_resources: bool = True,
):
    if not (model_ids or render_all):
        raise ValueError("At least one model id must be provided if --all is missing")

    site_url = URL(site_url_str or "")

    set_default_env()
    props, pm, _, plugin = load_common_data(props_file, branch="")

    with ModResourceLoader.load_all(props, pm, export=export_resources) as loader:
        if model_ids:
            with BlockRenderer(loader=loader, output_dir=output_dir) as renderer:
                for model_id in model_ids:
                    model_id = ResourceLocation.from_str(model_id)
                    render_block(model_id, renderer, site_url)
        else:
            asset_loader = plugin.asset_loader(
                loader=loader,
                site_url=site_url,
                asset_url=props.env.asset_url,
                render_dir=output_dir,
            )
            render_textures_and_export_metadata(loader, asset_loader)

    logger.info("Done.")


@app.command(deprecated=True)
def export(
    output_dir: PathArgument = DEFAULT_MERGE_SRC,
    *,
    branch: BranchOption,
    release: ReleaseOption = False,
    props_file: PropsOption,
):
    logger.warning("This command is deprecated, use `hexdoc build` instead.")
    build(
        output_dir,
        branch=branch,
        release=release,
        props_file=props_file,
    )


@app.command(deprecated=True)
def render(
    output_dir: PathArgument = DEFAULT_MERGE_SRC,
    *,
    branch: BranchOption,
    release: ReleaseOption = False,
    clean: bool = False,
    lang: Optional[str] = None,
    props_file: PropsOption,
):
    logger.warning("This command is deprecated, use `hexdoc build` instead.")
    if lang is not None:
        logger.warning(
            "`--lang` is deprecated and has been removed from `hexdoc build`."
        )
    build(
        output_dir,
        branch=branch,
        release=release,
        clean=clean,
        props_file=props_file,
    )


if __name__ == "__main__":
    app()



"""
A [Jinja](https://jinja.palletsprojects.com/en/3.1.x/)-based documentation generator for [Patchouli](https://github.com/VazkiiMods/Patchouli) books.
"""

from .__version__ import VERSION as VERSION

HEXDOC_MODID = "hexdoc"



__all__ = [
    "DefaultTyper",
    "init_context",
    "load_common_data",
    "render_textures_and_export_metadata",
]

from .args import (
    DefaultTyper,
)
from .load import (
    init_context,
    load_common_data,
    render_textures_and_export_metadata,
)



from contextlib import contextmanager
from contextvars import ContextVar
from typing import TypeVar

_T = TypeVar("_T")


@contextmanager
def set_contextvar(contextvar: ContextVar[_T], value: _T):
    token = contextvar.set(value)
    try:
        yield
    finally:
        contextvar.reset(token)



__all__ = [
    "BlastingRecipe",
    "CampfireCookingRecipe",
    "CookingRecipe",
    "CraftingRecipe",
    "CraftingShapedRecipe",
    "CraftingShapelessRecipe",
    "ItemIngredient",
    "ItemIngredientList",
    "ItemResult",
    "MinecraftItemIdIngredient",
    "MinecraftItemTagIngredient",
    "Recipe",
    "SmeltingRecipe",
    "SmithingRecipe",
    "SmithingTransformRecipe",
    "SmithingTrimRecipe",
    "SmokingRecipe",
    "StonecuttingRecipe",
]

from .ingredients import (
    ItemIngredient,
    ItemIngredientList,
    ItemResult,
    MinecraftItemIdIngredient,
    MinecraftItemTagIngredient,
)
from .recipes import (
    BlastingRecipe,
    CampfireCookingRecipe,
    CookingRecipe,
    CraftingRecipe,
    CraftingShapedRecipe,
    CraftingShapelessRecipe,
    Recipe,
    SmeltingRecipe,
    SmithingRecipe,
    SmithingTransformRecipe,
    SmithingTrimRecipe,
    SmokingRecipe,
    StonecuttingRecipe,
)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Mapping, TypeVar

if TYPE_CHECKING:
    from hexdoc.core import ModResourceLoader, ResourceLocation
    from hexdoc.utils import ContextSource, JSONDict


_Book = TypeVar("_Book")


class BookPlugin(ABC, Generic[_Book]):
    @property
    @abstractmethod
    def modid(self) -> str:
        """The modid of the mod whose book system this plugin implements."""

    @abstractmethod
    def load_book_data(
        self,
        book_id: ResourceLocation,
        loader: ModResourceLoader,
    ) -> tuple[ResourceLocation, JSONDict]:
        """"""

    @abstractmethod
    def is_i18n_enabled(self, book_data: Mapping[str, Any]) -> bool:
        """Given the raw book data, returns `True` if i18n is enabled for that book."""

    @abstractmethod
    def validate_book(
        self,
        book_data: Mapping[str, Any],
        *,
        context: ContextSource,
    ) -> _Book:
        """"""



import logging
from pathlib import Path

from .logging import TRACE

logger = logging.getLogger(__name__)


def strip_suffixes(path: Path) -> Path:
    """Removes all suffixes from a path. This is helpful because `path.with_suffix("")`
    only removes the last suffix.

    For example:
    ```py
    path = Path("lang/en_us.flatten.json5")
    strip_suffixes(path)  # lang/en_us
    path.with_suffix("")  # lang/en_us.flatten
    ```
    """
    while path.suffix:
        path = path.with_suffix("")
    return path


def replace_suffixes(path: Path, suffix: str) -> Path:
    """Replaces all suffixes of a path. This is helpful because `path.with_suffix()`
    only replaces the last suffix.

    For example:
    ```py
    path = Path("lang/en_us.flatten.json5")
    replace_suffixes(path, ".json")  # lang/en_us.json
    path.with_suffix(".json")        # lang/en_us.flatten.json
    ```
    """
    return strip_suffixes(path).with_suffix(suffix)


def write_to_path(path: Path, data: str | bytes, encoding: str = "utf-8"):
    logger.log(TRACE, f"Writing to {path}")
    path.parent.mkdir(parents=True, exist_ok=True)
    match data:
        case str():
            path.write_text(data, encoding)
        case _:
            path.write_bytes(data)



# pyright: reportPrivateUsage=false

from __future__ import annotations

import logging
import re
from enum import Enum, auto
from fnmatch import fnmatch
from typing import Literal, Self, final

from jinja2 import pass_context
from jinja2.runtime import Context
from pydantic import ValidationInfo, model_validator
from pydantic.dataclasses import dataclass
from pydantic.functional_validators import ModelWrapValidatorHandler

from hexdoc.core import Properties, ResourceLocation
from hexdoc.minecraft import I18n, LocalizedStr
from hexdoc.model import DEFAULT_CONFIG, HexdocModel, ValidationContextModel
from hexdoc.plugin import PluginManager
from hexdoc.utils import PydanticURL, TryGetEnum, classproperty
from hexdoc.utils.json_schema import inherited, json_schema_extra_config, type_str

logger = logging.getLogger(__name__)


DEFAULT_MACROS = {
    "$(obf)": "$(k)",
    "$(bold)": "$(l)",
    "$(strike)": "$(m)",
    "$(italic)": "$(o)",
    "$(italics)": "$(o)",
    "$(list": "$(li",
    "$(reset)": "$()",
    "$(clear)": "$()",
    "$(2br)": "$(br2)",
    "$(p)": "$(br2)",
    "/$": "$()",
    "<br>": "$(br)",
    "$(nocolor)": "$(0)",
    "$(item)": "$(#b0b)",
    "$(thing)": "$(#490)",
}

_REPLACEMENTS = {
    "br": "\n",
    "playername": "[Playername]",
}

_COLORS = {
    "0": "000",
    "1": "00a",
    "2": "0a0",
    "3": "0aa",
    "4": "a00",
    "5": "a0a",
    "6": "fa0",
    "7": "aaa",
    "8": "555",
    "9": "55f",
    "a": "5f5",
    "b": "5ff",
    "c": "f55",
    "d": "f5f",
    "e": "ff5",
    "f": "fff",
}

BookLinks = dict[str, PydanticURL]


class FormattingContext(ValidationContextModel):
    book_id: ResourceLocation
    macros: dict[str, str]


class BookLink(HexdocModel):
    raw_value: str
    id: ResourceLocation
    anchor: str | None

    @classmethod
    def from_str(cls, raw_value: str, book_id: ResourceLocation) -> Self:
        # anchor
        if "#" in raw_value:
            id_str, anchor = raw_value.split("#", 1)
        else:
            id_str, anchor = raw_value, None

        # case-insensitive id of the category or entry being linked to
        # namespace defaults to the namespace of the book, not minecraft
        id_str = id_str.lower()
        if ":" in id_str:
            id = ResourceLocation.from_str(id_str)
        else:
            id = book_id.with_path(id_str)

        return cls(raw_value=raw_value, id=id, anchor=anchor)

    @property
    def as_tuple(self) -> tuple[ResourceLocation, str | None]:
        return (self.id, self.anchor)

    @property
    def book_links_key(self) -> str:
        key = str(self.id)
        if self.anchor is not None:
            key += f"#{self.anchor}"
        return key

    @property
    def fragment(self) -> str:
        return f"#{self.raw_value.replace('#', '@')}"


# Higgledy piggledy
# Old fuck Alwinfy said,
# "Eschew your typechecks and
# live with a pair,"
#
# Making poor Object do
# Re-re-re-factoring
# Till Winfy took up her
# Classical flair.


class CommandStyleType(TryGetEnum):
    """Command styles, like `$(type)`."""

    obfuscated = "k"
    bold = "l"
    strikethrough = "m"
    underline = "n"
    italic = "o"

    @classproperty
    @classmethod
    def macro_group(cls) -> str:
        return "command"


class FunctionStyleType(TryGetEnum):
    """Function styles, like `$(type:value)`."""

    tooltip = "t"
    cmd_click = "c"

    @classproperty
    @classmethod
    def macro_group(cls) -> str:
        return "function"


class SpecialStyleType(Enum):
    """Styles with no defined name, like `$(#0080ff)`, or styles which must be handled
    differently than the normal styles, like `$()`."""

    base = auto()
    paragraph = auto()
    color = auto()
    link = "l"

    @classproperty
    @classmethod
    def macro_group(cls) -> str:
        return "special"


class Style(HexdocModel, frozen=True):
    type: CommandStyleType | FunctionStyleType | SpecialStyleType

    @staticmethod
    def parse(
        style_str: str,
        book_id: ResourceLocation,
        i18n: I18n,
        is_0_black: bool,
        link_overrides: dict[str, str],
    ) -> Style | _CloseTag | str:
        # direct text replacements
        if style_str in _REPLACEMENTS:
            return _REPLACEMENTS[style_str]

        # paragraph
        if style := ParagraphStyle.try_parse(style_str):
            return style

        # commands
        if style_type := CommandStyleType.get(style_str):
            return CommandStyle(type=style_type)

        # reset color, but only if 0 is considered reset instead of black
        if style_str == "0" and not is_0_black:
            return _CloseTag(type=SpecialStyleType.color)

        # preset colors
        if style_str in _COLORS:
            return FunctionStyle(type=SpecialStyleType.color, value=_COLORS[style_str])

        # hex colors (#rgb and #rrggbb)
        if style_str.startswith("#") and len(style_str) in [4, 7]:
            return FunctionStyle(type=SpecialStyleType.color, value=style_str[1:])

        # functions
        if ":" in style_str:
            name, value = style_str.split(":", 1)

            # keys
            if name == "k":
                return str(i18n.localize_key(value))

            # links
            if name == SpecialStyleType.link.value:
                return LinkStyle.from_str(value, book_id, link_overrides)

            # all the other functions
            if style_type := FunctionStyleType.get(name):
                return FunctionStyle(type=style_type, value=value)

        # reset
        if style_str == "":
            return _CloseTag(type=SpecialStyleType.base)

        # close functions
        if style_str.startswith("/"):
            # links
            if style_str[1:] == SpecialStyleType.link.value:
                return _CloseTag(type=SpecialStyleType.link)

            # all the other functions
            if style_type := FunctionStyleType.get(style_str[1:]):
                return _CloseTag(type=style_type)

        # oopsies
        raise ValueError(f"Unhandled style: {style_str}")

    @property
    def macro(self) -> str:
        return f"{self.type.macro_group}_{self.type.name}"


def is_external_link(value: str) -> bool:
    return value.startswith(("https:", "http:"))


class CommandStyle(Style, frozen=True):
    type: CommandStyleType | Literal[SpecialStyleType.base]


class ParagraphStyleSubtype(Enum):
    paragraph = auto()
    list_item = auto()


class ParagraphStyle(Style, frozen=True):
    type: Literal[SpecialStyleType.paragraph] = SpecialStyleType.paragraph
    subtype: ParagraphStyleSubtype

    @classmethod
    def try_parse(cls, style_str: str) -> ParagraphStyle | None:
        if style_str == "br2":
            return cls.paragraph()

        # https://github.com/VazkiiMods/Patchouli/blob/4522fbb3e4/Xplat/src/main/java/vazkii/patchouli/client/book/text/BookTextParser.java#L346-L355
        if re.fullmatch(r"li\d?", style_str):
            level_str = style_str.removeprefix("li")
            level = int(level_str) if level_str.isnumeric() else 1
            return ListItemStyle(level=level)

    @classmethod
    def paragraph(cls):
        return ParagraphStyle(subtype=ParagraphStyleSubtype.paragraph)

    @property
    def macro(self) -> str:
        return f"paragraph_{self.subtype.name}"


class ListItemStyle(ParagraphStyle, frozen=True):
    subtype: Literal[ParagraphStyleSubtype.list_item] = ParagraphStyleSubtype.list_item
    level: int


class FunctionStyle(Style, frozen=True):
    type: FunctionStyleType | Literal[SpecialStyleType.color]
    value: str


class LinkStyle(Style, frozen=True):
    type: Literal[SpecialStyleType.link] = SpecialStyleType.link
    value: str | BookLink
    external: bool

    @classmethod
    def from_str(
        cls,
        raw_value: str,
        book_id: ResourceLocation,
        link_overrides: dict[str, str],
    ) -> Self:
        value = raw_value
        external = False

        for link, override in link_overrides.items():
            if fnmatch(value, link):
                value = override
                break

        if is_external_link(value):
            external = True
        elif not value.startswith("?"):  # TODO: support query params in BookLink
            value = BookLink.from_str(value, book_id)

        return cls(value=value, external=external)

    @pass_context
    def href(self, context: Context | dict[{"book_links": BookLinks}]):  # noqa
        match self.value:
            case str(href):
                return href
            case BookLink(book_links_key=key) as book_link:
                book_links: BookLinks = context["book_links"]
                if key not in book_links:
                    logger.debug(f"{key=}\n{book_link=}\n{book_links=}")
                    raise ValueError(f"broken link: {book_link}")
                return str(book_links[key])


# intentionally not inheriting from Style, because this is basically an implementation
# detail of the parser and should not be returned or exposed anywhere
class _CloseTag(HexdocModel, frozen=True):
    type: (
        FunctionStyleType
        | Literal[
            SpecialStyleType.link,
            SpecialStyleType.base,
            SpecialStyleType.color,
        ]
    )


STYLE_REGEX = re.compile(r"\$\(([^)]*)\)")


@final
@dataclass(config=DEFAULT_CONFIG | json_schema_extra_config(type_str, inherited))
class FormatTree:
    style: Style
    children: list[FormatTree | str]  # this can't be Self, it breaks Pydantic
    raw: str | None = None

    @classmethod
    def format(
        cls,
        string: str,
        *,
        book_id: ResourceLocation,
        i18n: I18n,
        macros: dict[str, str],
        is_0_black: bool,
        pm: PluginManager,
        link_overrides: dict[str, str],
    ) -> Self:
        for macro, replace in macros.items():
            if macro in replace:
                raise RuntimeError(
                    f"Recursive macro: replacement `{replace}` is matched by key `{macro}`"
                )

        working_string = resolve_macros(string, macros)

        # lex out parsed styles
        text_nodes: list[str] = []
        styles: list[Style | _CloseTag] = []
        text_since_prev_style: list[str] = []
        last_end = 0

        for match in re.finditer(STYLE_REGEX, working_string):
            # get the text between the previous match and here
            leading_text = working_string[last_end : match.start()]
            text_since_prev_style.append(leading_text)
            last_end = match.end()

            match Style.parse(match[1], book_id, i18n, is_0_black, link_overrides):
                case str(replacement):
                    # str means "use this instead of the original value"
                    text_since_prev_style.append(replacement)
                case Style() | _CloseTag() as style:
                    # add this style and collect the text since the previous one
                    styles.append(style)
                    text_nodes.append("".join(text_since_prev_style))
                    text_since_prev_style.clear()

        text_nodes.append("".join(text_since_prev_style) + working_string[last_end:])
        first_node = text_nodes.pop(0)

        # parse
        style_stack = [
            FormatTree(CommandStyle(type=SpecialStyleType.base), []),
            FormatTree(ParagraphStyle.paragraph(), [first_node]),
        ]
        for style, text in zip(styles, text_nodes):
            tmp_stylestack: list[Style] = []
            if style.type == SpecialStyleType.base:
                while style_stack[-1].style.type != SpecialStyleType.paragraph:
                    last_node = style_stack.pop()
                    style_stack[-1].children.append(last_node)
            elif any(tree.style.type == style.type for tree in style_stack):
                while len(style_stack) >= 2:
                    last_node = style_stack.pop()
                    style_stack[-1].children.append(last_node)
                    if last_node.style.type == style.type:
                        break
                    tmp_stylestack.append(last_node.style)

            for sty in tmp_stylestack:
                style_stack.append(FormatTree(sty, []))

            if isinstance(style, _CloseTag):
                if text:
                    style_stack[-1].children.append(text)
            else:
                style_stack.append(FormatTree(style, [text] if text else []))

        while len(style_stack) >= 2:
            last_node = style_stack.pop()
            style_stack[-1].children.append(last_node)

        unvalidated_tree = style_stack[0]
        unvalidated_tree.raw = string

        validated_tree = pm.validate_format_tree(
            tree=unvalidated_tree,
            macros=macros,
            book_id=book_id,
            i18n=i18n,
            is_0_black=is_0_black,
            link_overrides=link_overrides,
        )
        assert isinstance(validated_tree, cls)

        return validated_tree

    @model_validator(mode="wrap")
    @classmethod
    def _wrap_root(
        cls,
        value: str | LocalizedStr | Self,
        handler: ModelWrapValidatorHandler[Self],
        info: ValidationInfo,
    ):
        if not info.context or isinstance(value, FormatTree):
            return handler(value)

        context = FormattingContext.of(info)
        i18n = I18n.of(info)
        pm = PluginManager.of(info)
        props = Properties.of(info)

        if isinstance(value, str):
            value = i18n.localize(value)

        return cls.format(
            value.value,
            book_id=context.book_id,
            i18n=i18n,
            macros=context.macros,
            is_0_black=props.is_0_black,
            pm=pm,
            link_overrides=props.link_overrides,
        )


def resolve_macros(string: str, macros: dict[str, str]) -> str:
    # this could use ahocorasick, but it works fine for now
    old_string = None
    while old_string != string:
        old_string = string
        for macro, replace in macros.items():
            string = string.replace(macro, replace)
    return string



import importlib
import logging
import os
from pathlib import Path
from typing import Annotated

from click import BadParameter
from typer import Argument, Option, Typer

from hexdoc.utils import commands
from hexdoc.utils.tracebacks import get_message_with_hints
from hexdoc.utils.types import typed_partial

logger = logging.getLogger(__name__)

DEFAULT_MERGE_SRC = Path("_site/src/docs")
DEFAULT_MERGE_DST = Path("_site/dst/docs")

DEFAULT_PROPS_FILES = [
    "hexdoc.toml",
    "doc/hexdoc.toml",
    "properties.toml",
    "doc/properties.toml",
]


DefaultTyper = typed_partial(Typer)(
    context_settings={
        "help_option_names": ["--help", "-h"],
    },
    add_completion=False,
    pretty_exceptions_show_locals=bool(os.getenv("HEXDOC_TYPER_EXCEPTION_LOCALS")),
)


def get_default_props() -> Path:
    for path in DEFAULT_PROPS_FILES:
        path = Path(path)
        if path.is_file():
            logger.debug(f"Loading props from default path: {path}")
            return path

    raise FileNotFoundError(
        f"Props file not found at any default path: {', '.join(DEFAULT_PROPS_FILES)}"
    )


def get_current_commit() -> str:
    return commands.run(["git", "rev-parse", "--short", "HEAD"])


PathArgument = Annotated[Path, Argument()]

ReleaseOption = Annotated[bool, Option(envvar="HEXDOC_RELEASE")]

VerbosityOption = Annotated[int, Option("--verbose", "-v", count=True)]

PropsOption = Annotated[
    Path,
    Option(
        "--props",
        "-p",
        envvar="HEXDOC_PROPS",
        default_factory=get_default_props,
    ),
]

BranchOption = Annotated[
    str,
    Option(
        "--branch",
        envvar="HEXDOC_BRANCH",
        default_factory=get_current_commit,
    ),
]


def parse_import_class(value: str):
    if "." not in value:
        raise BadParameter("import path must contain at least one '.' character.")
    module_name, attr_name = value.rsplit(".", 1)

    try:
        module = importlib.import_module(module_name, package="hexdoc")
    except ModuleNotFoundError as e:
        raise BadParameter(str(e))

    try:
        type_ = getattr(module, attr_name)
    except AttributeError as e:
        raise BadParameter(get_message_with_hints(e))

    if not isinstance(type_, type):
        raise BadParameter(f"{type_} is not a class.")

    return type_



import functools
from typing import Any, Callable, ParamSpec, TypeVar, cast

from jinja2 import pass_context
from jinja2.runtime import Context
from markupsafe import Markup

from hexdoc.core import Properties, ResourceLocation
from hexdoc.core.resource import ItemStack
from hexdoc.minecraft import I18n
from hexdoc.minecraft.assets import (
    ItemWithTexture,
    PNGTexture,
    validate_texture,
)
from hexdoc.patchouli import FormatTree
from hexdoc.plugin import PluginManager

_P = ParamSpec("_P")
_R = TypeVar("_R")


def make_jinja_exceptions_suck_a_bit_less(f: Callable[_P, _R]) -> Callable[_P, _R]:
    @functools.wraps(f)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            args_ = list(args)
            if args_ and isinstance(args_[0], Context):
                args_ = args_[1:]

            e.add_note(f"args:   {args_}")
            e.add_note(f"kwargs: {kwargs}")
            raise

    return wrapper


# filters


@make_jinja_exceptions_suck_a_bit_less
def hexdoc_wrap(value: str, *args: str):
    tag, *attributes = args
    if attributes:
        attributes = " " + " ".join(attributes)
    else:
        attributes = ""
    return Markup(f"<{tag}{attributes}>{Markup.escape(value)}</{tag}>")


# aliased as _() and _f() at render time
@make_jinja_exceptions_suck_a_bit_less
def hexdoc_localize(
    key: str,
    *,
    do_format: bool,
    props: Properties,
    book_id: ResourceLocation,
    i18n: I18n,
    macros: dict[str, str],
    pm: PluginManager,
):
    # get the localized value from i18n
    localized = i18n.localize(key)

    if not do_format:
        return Markup(localized.value)

    # construct a FormatTree from the localized value (to allow using patchi styles)
    formatted = FormatTree.format(
        localized.value,
        book_id=book_id,
        i18n=i18n,
        macros=macros,
        is_0_black=props.is_0_black,
        pm=pm,
        link_overrides=props.link_overrides,
    )
    return formatted


# TODO: support the full texture lookup
@pass_context
@make_jinja_exceptions_suck_a_bit_less
def hexdoc_texture(context: Context, id: str | ResourceLocation) -> str:
    texture = validate_texture(
        id,
        context=context,
        model_type=PNGTexture,
    )
    return str(texture.url)


@pass_context
@make_jinja_exceptions_suck_a_bit_less
def hexdoc_item(
    context: Context,
    id: str | ResourceLocation | ItemStack,
) -> ItemWithTexture:
    return ItemWithTexture.model_validate(
        id,
        context=cast(dict[str, Any], context),  # lie
    )


@pass_context
@make_jinja_exceptions_suck_a_bit_less
def hexdoc_smart_var(context: Context, value: Any):
    """Smart template argument filter.

    If `value` is of the form `{"variable": str(ref)}`, returns the value of the
    template variable called `ref`.

    Otherwise, returns `value` unchanged.
    """

    match value:
        case {**items} if len(items) != 1:
            return value
        case {"variable": str(ref)}:
            return context.resolve(ref)
        case _:
            return value



import re
from typing import TypeVar, cast

import pyjson5
from pydantic import JsonValue as JSONValue

from .assertions import isinstance_or_raise

_T_co = TypeVar("_T_co", covariant=True)

JSONDict = dict[str, JSONValue]


def decode_json_dict(data: str | bytes) -> JSONDict:
    match data:
        case str():
            decoded = pyjson5.decode(data)
        case _:
            decoded = pyjson5.decode_utf8(data)
    assert isinstance_or_raise(decoded, dict)
    return cast(JSONDict, decoded)


# implement pkpcpbp's flattening in python
# https://github.com/gamma-delta/PKPCPBP/blob/786194a590f/src/main/java/at/petrak/pkpcpbp/filters/JsonUtil.java
def decode_and_flatten_json_dict(data: str) -> dict[str, str]:
    # replace `\<LF>       foobar` with `\<LF>foobar`
    data = re.sub(r"\\\n\s*", "\\\n", data)

    # decode and flatten
    decoded = decode_json_dict(data)
    return _flatten_inner(decoded, "")


def _flatten_inner(obj: JSONDict, prefix: str) -> dict[str, str]:
    out: dict[str, str] = {}

    for key_stub, value in obj.items():
        if not prefix:
            key = key_stub
        elif not key_stub:
            key = prefix
        elif prefix[-1] in ":_-/":
            key = prefix + key_stub
        else:
            key = f"{prefix}.{key_stub}"

        match value:
            case dict():
                _update_disallow_duplicates(out, _flatten_inner(value, key))
            case str():
                _update_disallow_duplicates(out, {key: value})
            case _:
                raise TypeError(value)

    return out


def _update_disallow_duplicates(base: dict[str, _T_co], new: dict[str, _T_co]):
    for key, value in new.items():
        if key in base:
            raise ValueError(f"Duplicate key {key}\nold=`{base[key]}`\nnew=`{value}`")
        base[key] = value



__all__ = [
    "BlockModel",
    "Blockstate",
    "ItemModel",
    "load_model",
]

from .block import BlockModel
from .blockstate import Blockstate
from .item import ItemModel
from .load import load_model



__all__ = [
    "Book",
    "BookContext",
    "Category",
    "Entry",
    "FormatTree",
    "FormattingContext",
    "page",
]

from . import page
from .book import Book
from .book_context import BookContext
from .category import Category
from .entry import Entry
from .text import FormattingContext, FormatTree



# pyright: reportPrivateUsage=false

from enum import Enum
from typing import Final, Literal


class NoValueType(Enum):
    """Type of NoValue, a singleton representing the value of a nonexistent dict key."""

    _token = 0

    def __str__(self):
        return "NoValue"

    def __bool__(self) -> Literal[False]:
        return False


NoValue: Final = NoValueType._token
"""A singleton (like None) representing the value of a nonexistent dict key."""


class InheritType(Enum):
    """Type of Inherit, a singleton representing a value that should be inherited."""

    _token = 0

    def __str__(self):
        return "Inherit"


Inherit: Final = InheritType._token
"""A singleton (like None) representing a value that should be inherited."""



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Annotated, Any, ClassVar, Generic, Protocol, TypeVar

from packaging.specifiers import SpecifierSet
from pydantic import GetCoreSchemaHandler, ValidationInfo
from pydantic_core import core_schema
from typing_extensions import override

from hexdoc.model.base import HexdocModel

_T = TypeVar("_T")

_T_ModelType = TypeVar("_T_ModelType", bound=type[HexdocModel])

_If = TypeVar("_If")
_Else = TypeVar("_Else")


class VersionSource(Protocol):
    @classmethod
    def get(cls) -> str | None:
        """Returns the current version."""
        ...

    @classmethod
    def matches(cls, specifier: str | SpecifierSet) -> bool:
        """Returns True if the current version matches the version_spec."""
        ...


class MinecraftVersion(VersionSource):
    MINECRAFT_VERSION: ClassVar[str | None] = None

    @override
    @classmethod
    def get(cls) -> str | None:
        return cls.MINECRAFT_VERSION

    @override
    @classmethod
    def matches(cls, specifier: str | SpecifierSet) -> bool:
        if isinstance(specifier, str):
            specifier = SpecifierSet(specifier)
        if (version := cls.get()) is None:
            return True
        return version in specifier


@dataclass(frozen=True)
class Versioned:
    """Base class for types which can behave differently based on a version source,
    which defaults to MinecraftVersion."""

    version_spec: str
    version_source: VersionSource = field(default=MinecraftVersion, kw_only=True)

    @property
    def is_current(self):
        return self.version_source.matches(self.version_spec)


@dataclass(frozen=True)
class IsVersion(Versioned):
    """Instances of this class are truthy if version_spec matches version_source, which
    defaults to MinecraftVersion.

    Can be used as a Pydantic validator annotation, which raises ValueError if
    version_spec doesn't match the current version. Use it like this:

    `Annotated[str, IsVersion(">=1.20")] | Annotated[None, IsVersion("<1.20")]`

    Can also be used as a class decorator for Pydantic models, which raises ValueError
    when validating the model if version_spec doesn't match the current version.
    Decorated classes must subclass HexdocModel (or HexdocBaseModel).
    """

    def __bool__(self):
        return self.is_current

    def __call__(self, cls: _T_ModelType) -> _T_ModelType:
        cls.__hexdoc_before_validator__ = self._model_validator
        return cls

    def __get_pydantic_core_schema__(
        self,
        source_type: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_before_validator_function(
            self._schema_validator,
            schema=handler(source_type),
        )

    def _schema_validator(self, value: Any):
        if self.is_current:
            return value
        raise ValueError(
            f"Expected version {self.version_spec}, got {self.version_source.get()}"
        )

    def _model_validator(self, cls: Any, value: Any, info: ValidationInfo):
        return self._schema_validator(value)


Before_1_19 = Annotated[_T, IsVersion("<1.19")]
"""Alias for `Annotated[_T, IsVersion("<1.19")]`."""
AtLeast_1_19 = Annotated[_T, IsVersion(">=1.19")]
"""Alias for `Annotated[_T, IsVersion(">=1.19")]`."""
After_1_19 = Annotated[_T, IsVersion(">1.19")]
"""Alias for `Annotated[_T, IsVersion("<1.19")]`."""

Before_1_20 = Annotated[_T, IsVersion("<1.20")]
"""Alias for `Annotated[_T, IsVersion("<1.20")]`."""
AtLeast_1_20 = Annotated[_T, IsVersion(">=1.20")]
"""Alias for `Annotated[_T, IsVersion(">=1.20")]`."""
After_1_20 = Annotated[_T, IsVersion(">1.20")]
"""Alias for `Annotated[_T, IsVersion("<1.20")]`."""


@dataclass(frozen=True)
class ValueIfVersion(Versioned, Generic[_If, _Else]):
    value_if: _If
    value_else: _Else

    def __call__(self) -> _If | _Else:
        if self.is_current:
            return self.value_if
        return self.value_else



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, ClassVar, Generic, Self, TypeVar, Unpack

from pydantic import ConfigDict, Field, model_validator
from pydantic.functional_validators import ModelWrapValidatorHandler
from typing_extensions import override

from hexdoc.core import ResourceLocation
from hexdoc.minecraft import LocalizedStr
from hexdoc.minecraft.recipe import Recipe
from hexdoc.model import TypeTaggedTemplate
from hexdoc.utils import Inherit, InheritType, NoValue, classproperty

from ..text import FormatTree
from ..utils import AdvancementSpoilered, Flagged

_T_Recipe = TypeVar("_T_Recipe", bound=Recipe)

_T_PageWithAccumulator = TypeVar(
    "_T_PageWithAccumulator",
    bound="PageWithAccumulator[Any]",
    contravariant=True,
)

_T_PageWithRecipeAccumulator = TypeVar(
    "_T_PageWithRecipeAccumulator",
    bound="PageWithRecipeAccumulator[Any]",
    contravariant=True,
)


class Page(TypeTaggedTemplate, AdvancementSpoilered, Flagged, type=None):
    """Base class for Patchouli page types.

    See: https://vazkiimods.github.io/Patchouli/docs/patchouli-basics/page-types
    """

    advancement: ResourceLocation | None = None
    anchor: str | None = None

    def __init_subclass__(
        cls,
        *,
        type: str | InheritType | None = Inherit,
        template_type: str | None = None,
        **kwargs: Unpack[ConfigDict],
    ) -> None:
        super().__init_subclass__(type=type, template_type=template_type, **kwargs)

    @classproperty
    @classmethod
    def type(cls) -> ResourceLocation | None:
        assert cls._type is not NoValue
        return cls._type

    @model_validator(mode="wrap")
    @classmethod
    def _pre_root(cls, value: str | Any, handler: ModelWrapValidatorHandler[Self]):
        match value:
            case str(text):
                # treat a plain string as a text page
                value = {"type": "patchouli:text", "text": text}
            case {"type": str(raw_type)} if ":" not in raw_type:
                # default to the patchouli namespace if not specified
                # see: https://github.com/VazkiiMods/Patchouli/blob/b87e91a5a08d/Xplat/src/main/java/vazkii/patchouli/client/book/ClientBookRegistry.java#L110
                value["type"] = f"patchouli:{raw_type}"
            case _:
                pass
        return handler(value)

    @classproperty
    @classmethod
    def template(cls) -> str:
        return cls.template_id.template_path("pages")

    def book_link_key(self, entry_key: str):
        """Key to look up this page in `BookContext.book_links`, or `None` if this page
        has no anchor."""
        if self.anchor is not None:
            return f"{entry_key}#{self.anchor}"

    def fragment(self, entry_fragment: str):
        """URL fragment for this page in `BookContext.book_links`, or `None` if this
        page has no anchor."""
        if self.anchor is not None:
            return f"{entry_fragment}@{self.anchor}"

    def redirect_path(self, entry_path: str):
        """Path to this page when generating redirect pages, or `None` if this page has
        no anchor."""
        if self.anchor is not None:
            return f"{entry_path}/{self.anchor}"

    def _get_advancement(self):
        # implements AdvancementSpoilered
        return self.advancement


class PageWithText(Page, type=None):
    """Base class for a `Page` with optional text.

    If text is required, do not subclass this type.
    """

    text: FormatTree | None = None


class PageWithTitle(PageWithText, type=None):
    """Base class for a `Page` with optional title and text.

    If title and/or text is required, do not subclass this type.
    """

    title: LocalizedStr | None = None


class PageWithDoubleRecipe(PageWithTitle, Generic[_T_Recipe], type=None):
    recipe: _T_Recipe
    recipe2: _T_Recipe | None = None

    @property
    def recipes(self) -> list[_T_Recipe]:
        return [r for r in [self.recipe, self.recipe2] if r is not None]


class PageWithAccumulator(Page, ABC, type=None):
    """Base class for pages that can merge together when adjacent."""

    @classmethod
    @abstractmethod
    def accumulator_type(cls) -> type[AccumulatorPage[Any]]:
        """Returns the RecipeAccumulator class for this page type.

        The template type of the returned class must match that of this class.
        """

    @property
    @abstractmethod
    def accumulator_title(self) -> LocalizedStr | None:
        """Returns the page's title, if any, for use in the accumulator."""

    @property
    @abstractmethod
    def accumulator_text(self) -> FormatTree | None:
        """Returns the page's text, if any, for use in the accumulator."""


class PageWithRecipeAccumulator(
    PageWithAccumulator, ABC, Generic[_T_Recipe], type=None
):
    @property
    @abstractmethod
    def accumulator_recipes(self) -> list[_T_Recipe]:
        """Returns the page's recipes for use in the accumulator."""


class PageWithDoubleRecipeAccumulator(
    PageWithDoubleRecipe[_T_Recipe],
    PageWithRecipeAccumulator[_T_Recipe],
    ABC,
    Generic[_T_Recipe],
    type=None,
):
    @property
    @override
    def accumulator_title(self) -> LocalizedStr | None:
        return self.title

    @property
    @override
    def accumulator_text(self) -> FormatTree | None:
        return self.text

    @property
    @override
    def accumulator_recipes(self) -> list[_T_Recipe]:
        return self.recipes


class AccumulatorPage(PageWithTitle, ABC, Generic[_T_PageWithAccumulator], type=None):
    """Base class for virtual pages generated to merge adjacent instances of a
    PageWithAccumulator page type together."""

    _page_type: ClassVar[ResourceLocation | None]

    def __init_subclass__(
        cls,
        *,
        page_type: type[PageWithAccumulator[_T_PageWithAccumulator]] | None = None,
        **kwargs: Unpack[ConfigDict],
    ) -> None:
        if page_type:
            super().__init_subclass__(
                type=None,
                template_type=str(page_type.template_id),
                **kwargs,
            )
            cls._page_type = page_type.type
        else:
            super().__init_subclass__(type=None, template_type=None, **kwargs)
            cls._page_type = None

    @classmethod
    def from_page(cls, page: _T_PageWithAccumulator) -> Self:
        """Constructs a new accumulator from the given page.

        Note: `append(page)` is always called immediately after this.
        """
        if cls._page_type is None:
            raise RuntimeError(f"Cannot instantiate {cls} because page_type is None")
        if page.type != cls._page_type:
            raise ValueError(f"Mismatched page type: {cls}, {page}")

        self = cls.model_construct()
        self.title = page.accumulator_title
        self.anchor = page.anchor
        self.advancement = page.advancement
        return self

    @property
    def has_content(self) -> bool:
        """Returns True if this accumulator contains any user-visible content."""
        return bool(self.title or self.text)

    def can_append(self, page: _T_PageWithAccumulator) -> bool:
        """Returns True if this accumulator can append the given page."""
        return (
            self._page_type == page.type
            and self.title == page.accumulator_title
            and self.anchor == page.anchor
            and self.advancement == page.advancement
        )

    @abstractmethod
    def append(self, page: _T_PageWithAccumulator):
        """Appends the given page to this accumulator."""
        self.text = page.accumulator_text

    @property
    def can_append_more(self) -> bool:
        """Returns True if this accumulator can append more pages."""
        return self.text is None


class RecipeAccumulatorPage(
    AccumulatorPage[_T_PageWithRecipeAccumulator],
    Generic[_T_PageWithRecipeAccumulator, _T_Recipe],
):
    recipes: list[_T_Recipe] = Field(default_factory=lambda: [])

    @property
    @override
    def has_content(self) -> bool:
        return super().has_content or bool(self.recipes)

    @override
    def append(self, page: _T_PageWithRecipeAccumulator):
        super().append(page)
        self.recipes += page.accumulator_recipes



# pyright: reportPrivateUsage=false

import os
import platform
from importlib.resources import Package
from pathlib import Path
from typing import Any, Mapping

import hexdoc
from hexdoc import HEXDOC_MODID, VERSION
from hexdoc.core import IsVersion, ModResourceLoader, ResourceLocation
from hexdoc.minecraft.recipe import (
    ingredients as minecraft_ingredients,
    recipes as minecraft_recipes,
)
from hexdoc.patchouli import Book, BookContext
from hexdoc.patchouli.page import pages as patchouli_pages
from hexdoc.plugin import (
    BookPlugin,
    BookPluginImpl,
    HookReturn,
    LoadTaggedUnionsImpl,
    ModPlugin,
    ModPluginImpl,
    hookimpl,
)
from hexdoc.plugin.mod_plugin import DefaultRenderedTemplates
from hexdoc.utils import ContextSource, JSONDict, cast_context


def redirect_path(raw_link: str):
    # colon is forbidden in filenames on Windows
    if platform.system() == "Windows" or os.getenv("MOCK_PLATFORM") == "Windows":
        raw_link = raw_link.replace(":", "-")

    return f"{raw_link}/index.html"


class HexdocPlugin(LoadTaggedUnionsImpl, ModPluginImpl, BookPluginImpl):
    @staticmethod
    @hookimpl
    def hexdoc_mod_plugin(branch: str) -> ModPlugin:
        return HexdocModPlugin(branch=branch)

    @staticmethod
    @hookimpl
    def hexdoc_book_plugin() -> BookPlugin[Any]:
        return PatchouliBookPlugin()

    @staticmethod
    @hookimpl
    def hexdoc_load_tagged_unions() -> HookReturn[Package]:
        return [
            patchouli_pages,
            minecraft_recipes,
            minecraft_ingredients,
        ]


class HexdocModPlugin(ModPlugin):
    @property
    def modid(self):
        return HEXDOC_MODID

    @property
    def full_version(self):
        return VERSION

    @property
    def plugin_version(self):
        return VERSION

    def resource_dirs(self) -> HookReturn[Package]:
        from hexdoc._export import generated, resources

        return [generated, resources]

    def jinja_template_root(self) -> tuple[Package, str] | None:
        return hexdoc, "_templates"

    def default_rendered_templates(self) -> dict[str | Path, str]:
        return {
            "index.html": "index.html.jinja",
            "index.css": "index.css.jinja",
            "textures.css": "textures.jcss.jinja",
            "index.js": "index.js.jinja",
        }

    def default_rendered_templates_v2(
        self,
        book: Any,
        context: ContextSource,
    ) -> DefaultRenderedTemplates:
        if not isinstance(book, Book):
            return {}

        templates: DefaultRenderedTemplates = {}

        for category in book.categories.values():
            templates[redirect_path(category.redirect_path)] = (
                "redirects/category.html.jinja",
                {
                    "category": category,
                    "fragment": category.fragment,
                },
            )

            for entry in category.entries.values():
                templates[redirect_path(entry.redirect_path)] = (
                    "redirects/entry.html.jinja",
                    {
                        "category": category,
                        "entry": entry,
                        "fragment": entry.fragment,
                    },
                )

                for page in entry.pages:
                    page_path = page.redirect_path(entry.redirect_path)
                    if page_path is not None:
                        templates[redirect_path(page_path)] = (
                            "redirects/page.html.jinja",
                            {
                                "category": category,
                                "entry": entry,
                                "page": page,
                                "fragment": page.fragment(entry.fragment),
                            },
                        )

        return templates


class PatchouliBookPlugin(BookPlugin[Book]):
    @property
    def modid(self):
        return "patchouli"

    def load_book_data(
        self,
        book_id: ResourceLocation,
        loader: ModResourceLoader,
    ) -> tuple[ResourceLocation, JSONDict]:
        _, data = loader.load_resource("data", "patchouli_books", book_id / "book")

        if IsVersion("<1.20") and "extend" in data:
            book_id = ResourceLocation.model_validate(data["extend"])
            return self.load_book_data(book_id, loader)

        return book_id, data

    def is_i18n_enabled(self, book_data: Mapping[str, Any]):
        return book_data.get("i18n", False) is True

    def validate_book(
        self,
        book_data: Mapping[str, Any],
        *,
        context: ContextSource,
    ):
        book_ctx = BookContext.of(context)
        loader = ModResourceLoader.of(context)

        book = Book.model_validate(book_data, context=cast_context(context))
        book._load_categories(context, book_ctx)
        book._load_entries(context, book_ctx, loader)

        return book



__all__ = [
    "METADATA_SUFFIX",
    "AssumeTag",
    "BaseProperties",
    "BaseResourceDir",
    "BaseResourceLocation",
    "BookFolder",
    "Entity",
    "ExportFn",
    "IsVersion",
    "ItemStack",
    "MinecraftVersion",
    "ModResourceLoader",
    "PathResourceDir",
    "PluginResourceDir",
    "Properties",
    "ResLoc",
    "ResourceDir",
    "ResourceLocation",
    "ResourceType",
    "ValueIfVersion",
    "VersionSource",
    "Versioned",
    "compat",
    "properties",
]

from .compat import (
    IsVersion,
    MinecraftVersion,
    ValueIfVersion,
    Versioned,
    VersionSource,
)
from .loader import (
    METADATA_SUFFIX,
    BookFolder,
    ExportFn,
    ModResourceLoader,
)
from .properties import BaseProperties, Properties
from .resource import (
    AssumeTag,
    BaseResourceLocation,
    Entity,
    ItemStack,
    ResLoc,
    ResourceLocation,
    ResourceType,
)
from .resource_dir import (
    BaseResourceDir,
    PathResourceDir,
    PluginResourceDir,
    ResourceDir,
)



__all__ = [
    "IncludeRawExtension",
    "hexdoc_item",
    "hexdoc_localize",
    "hexdoc_smart_var",
    "hexdoc_texture",
    "hexdoc_wrap",
]

from .extensions import IncludeRawExtension
from .filters import (
    hexdoc_item,
    hexdoc_localize,
    hexdoc_smart_var,
    hexdoc_texture,
    hexdoc_wrap,
)



"""Generate common favicon formats from a single source image."""

# Standard Library
import asyncio
import json as _json
import math
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    Callable,
    Collection,
    Coroutine,
    Generator,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypedDict,
    Union,
)

# Third Party
from PIL import Image as PILImage

from ._constants import HTML_LINK, SUPPORTED_FORMATS
from ._exceptions import FaviconNotSupportedError
from ._types import Color, FaviconProperties
from ._util import generate_icon_types, svg_to_png, validate_path

LoosePath = Union[Path, str]
LooseColor = Union[Collection[int], str]


class FaviconDict(TypedDict):
    image_format: str
    dimensions: tuple[int, int]
    prefix: str
    rel: str | None


class Favicons:
    """Generate common favicon formats from a single source image."""

    def __init__(
        self,
        source: LoosePath,
        output_directory: LoosePath,
        background_color: LooseColor = "#000000",
        transparent: bool = True,
        base_url: str = "/",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize Favicons class."""
        self._validated = False
        self._output_directory = output_directory
        self._formats = tuple(generate_icon_types())
        self.transparent = transparent
        self.base_url = base_url
        self.background_color: Color = Color(background_color)
        self.generate: Union[Callable, Coroutine] = self.sgenerate
        self.completed: int = 0
        self._temp_source: Optional[Path] = None

        if isinstance(source, str):
            source = Path(source)

        self._source = source

        self._check_source_format()

    def _validate(self) -> None:
        self.source = validate_path(self._source)
        self.output_directory = validate_path(self._output_directory, create=True)

        if self.source.suffix.lower() not in SUPPORTED_FORMATS:
            raise FaviconNotSupportedError(self.source)

        self._validated = True

    def __enter__(self) -> "Favicons":
        """Enter Favicons context."""
        self._validate()
        self.generate = self.sgenerate
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        """Exit Favicons context."""
        self._close_temp_source()
        pass

    async def __aenter__(self) -> "Favicons":
        """Enter Favicons context."""
        self._validate()
        self.generate = self.agenerate
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        """Exit Favicons context."""
        self._close_temp_source()
        pass

    def _close_temp_source(self) -> None:
        """Close temporary file if it exists."""
        if self._temp_source is not None:
            try:
                self._temp_source.unlink()
            except FileNotFoundError:
                pass

    def _check_source_format(self) -> None:
        """Convert source image to PNG if it's in SVG format."""
        if self._source.suffix == ".svg":
            self._source = svg_to_png(self._source, self.background_color)

    @staticmethod
    def _get_center_point(background: PILImage, foreground: PILImage) -> Tuple:
        """Generate a tuple of center points for PIL."""
        bg_x, bg_y = background.size[0:2]
        fg_x, fg_y = foreground.size[0:2]
        x1 = math.floor((bg_x / 2) - (fg_x / 2))
        y1 = math.floor((bg_y / 2) - (fg_y / 2))
        x2 = math.floor((bg_x / 2) + (fg_x / 2))
        y2 = math.floor((bg_y / 2) + (fg_y / 2))
        return (x1, y1, x2, y2)

    def _generate_single(self, format_properties: FaviconProperties) -> None:
        with PILImage.open(self.source) as src:
            output_file = self.output_directory / str(format_properties)
            bg: Tuple[int, ...] = self.background_color.colors

            # If transparency is enabled, add alpha channel to color.
            if self.transparent:
                bg += (0,)

            # Create background.
            dst = PILImage.new("RGBA", format_properties.dimensions, bg)

            # Resize source image without changing aspect ratio.
            src.thumbnail(format_properties.dimensions)

            # Place source image on top of background image.
            dst.paste(src, box=self._get_center_point(dst, src))

            # Save new file.
            dst.save(output_file, format_properties.image_fmt)

            self.completed += 1

    async def _agenerate_single(self, format_properties: FaviconProperties) -> None:
        """Awaitable version of _generate_single."""

        return self._generate_single(format_properties)

    def sgenerate(self) -> None:
        """Generate favicons."""
        if not self._validated:
            self._validate()

        for fmt in self._formats:
            self._generate_single(fmt)

    async def agenerate(self) -> None:
        """Generate favicons."""
        if not self._validated:
            self._validate()

        await asyncio.gather(*(self._agenerate_single(fmt) for fmt in self._formats))

    def html_gen(self) -> Iterator[str]:
        """Get generator of HTML strings."""
        for fmt in self._formats:
            yield HTML_LINK.format(
                rel=fmt.rel,
                type=f"image/{fmt.image_fmt}",
                href=self.base_url + str(fmt),
            )

    def html(self) -> tuple[str, ...]:
        """Get tuple of HTML strings."""
        return tuple(self.html_gen())

    def formats(self):
        """Get image formats as list."""
        return tuple(f.dict() for f in self._formats)

    def json(self, *args: Any, **kwargs: Any) -> str:
        """Get image formats as JSON string."""
        return _json.dumps(self.formats(), *args, **kwargs)

    def filenames_gen(self, prefix: bool = False) -> Generator:
        """Get generator of favicon file names."""
        for fmt in self._formats:
            filename = str(fmt)
            if prefix:
                filename = self.base_url + filename
            yield filename

    def filenames(self, prefix: bool = False) -> Tuple:
        """Get tuple of favicon file names."""
        return tuple(self.filenames_gen(prefix=prefix))



from __future__ import annotations

from typing import Self

from yarl import URL

from hexdoc.core import ItemStack, ResourceLocation
from hexdoc.utils import ContextSource

from .animated import AnimatedTexture
from .textures import BaseTexture, PNGTexture

ImageTexture = PNGTexture | AnimatedTexture


# this needs to be a separate class, rather than just using ImageTexture directly,
# because the key in the lookup for SingleItemTexture is the item id, not the texture id
class SingleItemTexture(BaseTexture):
    inner: ImageTexture

    @classmethod
    def from_url(cls, url: URL, pixelated: bool) -> Self:
        return cls(
            inner=PNGTexture(url=url, pixelated=pixelated),
        )

    @classmethod
    def load_id(cls, id: ResourceLocation | ItemStack, context: ContextSource):
        return super().load_id(id.id, context)

    @property
    def url(self):
        return self.inner.url


class MultiItemTexture(BaseTexture):
    inner: list[ImageTexture]
    gaslighting: bool

    @classmethod
    def from_url(cls, url: URL, pixelated: bool) -> Self:
        return cls(
            inner=[PNGTexture(url=url, pixelated=pixelated)],
            gaslighting=False,
        )

    @classmethod
    def load_id(cls, id: ResourceLocation | ItemStack, context: ContextSource):
        return super().load_id(id.id, context)


ItemTexture = SingleItemTexture | MultiItemTexture



# pyright: reportUnknownArgumentType=false, reportUnknownMemberType=false

from __future__ import annotations

import shutil
from abc import ABC, abstractmethod
from contextlib import ExitStack, contextmanager
from fnmatch import fnmatch
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, ContextManager, Iterable, Iterator, Literal, Self
from zipfile import ZipFile

import importlib_resources as resources
from pydantic import Field, model_validator
from typing_extensions import override

from hexdoc.model import HexdocModel
from hexdoc.model.base import DEFAULT_CONFIG
from hexdoc.plugin import PluginManager
from hexdoc.utils import JSONDict, RelativePath
from hexdoc.utils.types import cast_nullable


class BaseResourceDir(HexdocModel, ABC):
    @staticmethod
    def _json_schema_extra(schema: dict[str, Any]):
        properties = schema.pop("properties")
        new_schema = {
            "anyOf": [
                schema | {"properties": properties | {key: value}}
                for key, value in {
                    "external": properties.pop("external"),
                    "internal": {
                        "type": "boolean",
                        "default": True,
                        "title": "Internal",
                    },
                }.items()
            ],
        }
        schema.clear()
        schema.update(new_schema)

    model_config = DEFAULT_CONFIG | {
        "json_schema_extra": _json_schema_extra,
    }

    external: bool
    reexport: bool
    """If not set, the default value will be `not self.external`.

    Must be defined AFTER `external` in the Pydantic model.
    """

    @abstractmethod
    def load(
        self,
        pm: PluginManager,
    ) -> ContextManager[Iterable[PathResourceDir]]: ...

    @property
    def internal(self):
        return not self.external

    @model_validator(mode="before")
    @classmethod
    def _default_reexport(cls, data: JSONDict | Any):
        if not isinstance(data, dict):
            return data

        external = cls._get_external(data)
        if external is None:
            return data

        if "reexport" not in data:
            data["reexport"] = not external

        return data

    @classmethod
    def _get_external(cls, data: JSONDict | Any):
        match data:
            case {"external": bool(), "internal": bool()}:
                raise ValueError(f"Expected internal OR external, got both: {data}")
            case {"external": bool(external)}:
                return external
            case {"internal": bool(internal)}:
                data.pop("internal")
                external = data["external"] = not internal
                return external
            case _:
                return None


class BasePathResourceDir(BaseResourceDir):
    # direct paths are probably internal
    external: bool = False
    reexport: bool = True

    required: bool = True
    """If false, ignore "file not found" errors."""

    @property
    @abstractmethod
    def _paths(self) -> Iterable[Path]: ...

    def _validate_path(self, path: Path) -> bool:
        return path.exists()

    @model_validator(mode="after")
    def _validate_paths(self):
        if self.required:
            for path in self._paths:
                if not self._validate_path(path):
                    raise ValueError(
                        f"{self.__class__.__name__} path does not exist: {path}"
                    )
        return self


class PathResourceDir(BasePathResourceDir):
    """Represents a path to a resources directory or a mod's `.jar` file."""

    @staticmethod
    def _json_schema_extra(schema: dict[str, Any]):
        BaseResourceDir._json_schema_extra(schema)
        new_schema = {
            "anyOf": [
                {
                    "type": "string",
                    "format": "path",
                },
                *schema["anyOf"],
            ]
        }
        schema.clear()
        schema.update(new_schema)

    model_config = DEFAULT_CONFIG | {
        "json_schema_extra": _json_schema_extra,
    }

    path: RelativePath
    """A path relative to `hexdoc.toml`."""

    archive: bool = Field(default=None, validate_default=False)  # type: ignore
    """If true, treat this path as a zip archive (eg. a mod's `.jar` file).

    If `path` ends with `.jar` or `.zip`, defaults to `True`.
    """

    # not a props field
    _modid: str | None = None

    @property
    def modid(self):
        return self._modid

    @property
    @override
    def _paths(self):
        return [self.path]

    def set_modid(self, modid: str) -> Self:
        self._modid = modid
        return self

    @contextmanager
    @override
    def load(self, pm: PluginManager):
        if self.archive:
            with self._extract_archive() as path:
                update = {
                    "path": path,
                    "archive": False,
                }
                yield [self.model_copy(update=update)]
        else:
            yield [self]

    @contextmanager
    def _extract_archive(self) -> Iterator[Path]:
        with (
            ZipFile(self.path, "r") as zf,
            TemporaryDirectory(suffix=self.path.name) as tempdir,
        ):
            # extract root-level files and *useful* sub-directories
            # ie. avoid extracting classes etc
            for info in zf.filelist:
                path = info.filename
                if path.startswith(("assets/", "data/")) or "/" not in path:
                    zf.extract(info, tempdir)

            yield Path(tempdir)

    @model_validator(mode="before")
    def _pre_root(cls: Any, value: Any):
        # treat plain strings as paths
        if isinstance(value, str):
            return {"path": value}
        return value

    @model_validator(mode="after")
    def _post_root(self):
        if cast_nullable(self.archive) is None:
            self.archive = self.path.suffix in {".jar", ".zip"}
        return self


class GlobResourceDir(BasePathResourceDir):
    glob: RelativePath
    exclude: list[str] = Field(default_factory=list)

    _resolved_paths: list[Path] | None = None

    @property
    @override
    def _paths(self):
        if self._resolved_paths:
            return self._resolved_paths

        base = Path()
        for i, part in enumerate(self.glob.parts):
            if "*" in part:
                glob = str(Path(*self.glob.parts[i:]))
                break
            base /= part
        else:
            raise ValueError(f"Glob does not contain any wildcards: {self.glob}")

        self._resolved_paths = [
            path
            for path in base.glob(glob)
            if not any(fnmatch(path.as_posix(), pat) for pat in self.exclude)
        ]
        if not self._resolved_paths:
            raise ValueError(f"Glob failed to find any matches: {self.glob}")

        return self._resolved_paths

    @contextmanager
    @override
    def load(self, pm: PluginManager):
        with ExitStack() as stack:
            yield [
                resource_dir
                for path in self._paths
                for resource_dir in stack.enter_context(
                    PathResourceDir(
                        path=path,
                        external=self.external,
                        reexport=self.reexport,
                    ).load(pm)
                )
            ]


class PatchouliBooksResourceDir(BasePathResourceDir):
    """For modpack books, eg. `.minecraft/patchouli_books`.

    Unlike other resource dirs, this points directly at `patchouli_books`, *not* at the
    `resources` directory. The namespace is always `patchouli`.

    https://vazkiimods.github.io/Patchouli/docs/patchouli-basics/getting-started#1-locate-patchouli_books
    """

    patchouli_books: RelativePath

    @property
    def modid(self) -> Literal["patchouli"]:
        return "patchouli"

    @property
    @override
    def _paths(self):
        return [self.patchouli_books]

    @contextmanager
    @override
    def load(self, pm: PluginManager):
        with TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)

            # TODO: hack
            # at this point, we don't know if it's a resource pack book or not
            # we would need to load book.json to figure that out
            # so just put it in both possible locations to cover all the bases
            for folder in ["assets", "data"]:
                dst = tmpdir / folder / self.modid / "patchouli_books"
                dst.parent.mkdir(parents=True)
                shutil.copytree(self.patchouli_books, dst)

            yield [
                PathResourceDir(
                    path=tmpdir,
                    external=self.external,
                    reexport=self.reexport,
                ).set_modid(self.modid)
            ]


class PluginResourceDir(BaseResourceDir):
    modid: str

    # if we're specifying a modid, it's probably from some other mod/package
    external: bool = True
    reexport: bool = False

    @contextmanager
    @override
    def load(self, pm: PluginManager):
        with ExitStack() as stack:
            yield list(self._load_all(pm, stack))  # NOT "yield from"

    def _load_all(self, pm: PluginManager, stack: ExitStack):
        for module in pm.load_resources(self.modid):
            traversable = resources.files(module)
            path = stack.enter_context(resources.as_file(traversable))

            yield PathResourceDir(
                path=path,
                external=self.external,
                reexport=self.reexport,
            ).set_modid(self.modid)  # setting _modid directly causes a validation error


ResourceDir = (
    PathResourceDir | PatchouliBooksResourceDir | PluginResourceDir | GlobResourceDir
)



from __future__ import annotations

import logging
from typing import Generic, TypeVar

from pydantic import field_validator

from hexdoc.core import (
    ItemStack,
    ResourceLocation,
)
from hexdoc.core.resource import BaseResourceLocation
from hexdoc.model import (
    HexdocModel,
    InlineItemModel,
    InlineModel,
)
from hexdoc.utils import ContextSource

from ..i18n import I18n, LocalizedStr
from .animated import AnimatedTexture
from .constants import TAG_TEXTURE_URL
from .items import ImageTexture, ItemTexture, MultiItemTexture, SingleItemTexture
from .load_assets import Texture
from .textures import PNGTexture

logger = logging.getLogger(__name__)

_T_BaseResourceLocation = TypeVar("_T_BaseResourceLocation", bound=BaseResourceLocation)

_T_Texture = TypeVar("_T_Texture", bound=Texture)


class BaseWithTexture(HexdocModel, Generic[_T_BaseResourceLocation, _T_Texture]):
    id: _T_BaseResourceLocation
    name: LocalizedStr
    texture: Texture

    @property
    def image_texture(self) -> ImageTexture | None:
        match self.texture:
            case PNGTexture() | AnimatedTexture():
                return self.texture
            case SingleItemTexture():
                return self.texture.inner
            case MultiItemTexture():
                return None

    @property
    def image_textures(self) -> list[ImageTexture] | None:
        match self.texture:
            case MultiItemTexture():
                return self.texture.inner
            case PNGTexture() | AnimatedTexture() | SingleItemTexture():
                return None

    @property
    def gaslighting(self) -> bool:
        match self.texture:
            case MultiItemTexture():
                return self.texture.gaslighting
            case PNGTexture() | AnimatedTexture() | SingleItemTexture():
                return False


class ItemWithTexture(InlineItemModel, BaseWithTexture[ItemStack, ItemTexture]):
    @classmethod
    def load_id(cls, item: ItemStack, context: ContextSource):
        """Implements InlineModel."""

        i18n = I18n.of(context)
        if (name := item.get_name()) is not None:
            pass
        elif item.path.startswith("texture"):
            name = i18n.localize_texture(item.id)
        else:
            name = i18n.localize_item(item)

        return {
            "id": item,
            "name": name,
            "texture": item.id,  # TODO: fix InlineModel (ItemTexture), then remove .id
        }


class TagWithTexture(InlineModel, BaseWithTexture[ResourceLocation, Texture]):
    @classmethod
    def load_id(cls, id: ResourceLocation, context: ContextSource):
        i18n = I18n.of(context)
        return cls(
            id=id,
            name=i18n.localize_item_tag(id),
            texture=PNGTexture.from_url(TAG_TEXTURE_URL, pixelated=True),
        )

    @field_validator("id", mode="after")
    @classmethod
    def _validate_id(cls, id: ResourceLocation):
        assert id.is_tag, f"Expected tag id, got {id}"
        return id


class NamedTexture(InlineModel, BaseWithTexture[ResourceLocation, ImageTexture]):
    @classmethod
    def load_id(cls, id: ResourceLocation, context: ContextSource):
        i18n = I18n.of(context)
        return {
            "id": id,
            "name": i18n.localize_texture(id, silent=True),
            "texture": id,
        }



"""Custom Validation & Conversion for Color."""

# Standard Library
import re
from typing import Collection, Generator, Pattern, Tuple, Union

# Project
from .._exceptions import FaviconColorError, FaviconsError

_RGB_STR_PATTERN = re.compile(r"^rgb\((\d+),\s?(\d+),\s?(\d+)\)$")
_HEX_STR_PATTERN3 = re.compile(r"^\#?([A-Fa-f0-9])([A-Fa-f0-9])([A-Fa-f0-9])$")
_HEX_STR_PATTERN6 = re.compile(r"^\#?([A-Fa-f0-9]{2})([A-Fa-f0-9]{2})([A-Fa-f0-9]{2})$")

Rgb = Tuple[int, int, int]


def get_matches(pattern: Pattern, search: str) -> Generator[str, None, None]:
    """Get capture group matches as a tuple."""
    matches = pattern.search(search)
    if matches is not None:
        for match in matches.groups():
            yield match
    else:
        raise FaviconsError(
            f"Pattern {repr(pattern)} did not match anything in {repr(search)}"
        )


class Color:
    """Validate & Transform Color Representations."""

    red: int
    green: int
    blue: int

    def __init__(self, color: Union[str, Collection]) -> None:
        """Validate & parse input color."""
        rgb = None
        if isinstance(color, str):
            rgb = self._validate_str(color)
        elif isinstance(color, Collection):
            rgb = self._validate_rgb_iter(color)
        else:
            raise FaviconColorError(color)
        if rgb is None:
            raise FaviconColorError(color)

        color_names = ("red", "green", "blue")
        for i, color_num in enumerate(rgb):
            setattr(self, color_names[i], color_num)

    def as_hex(self) -> str:
        """Represent color as hex value."""

        def hexstr(num: int) -> str:
            return hex(num).replace("0x", "").zfill(2)

        colors = (hexstr(c) for c in (self.red, self.green, self.blue))

        return "#" + "".join(colors)

    def as_rgb(self) -> str:
        """Represent color as RGB string in rgb(x, y, z) format."""
        return f"rgb({', '.join(str(c) for c in (self.red, self.green, self.blue))})"

    @property
    def colors(self) -> Rgb:
        """Represent color as tuple of RGB values."""
        return (self.red, self.green, self.blue)

    def _validate_str(self, colorstr: str) -> Generator:
        """Validate & parse input string."""
        rgb = None

        if _RGB_STR_PATTERN.match(colorstr):
            rgb = self._parse_rgbstr(colorstr)

        elif _HEX_STR_PATTERN3.match(colorstr):
            rgb = self._parse_hex(colorstr, 3)

        elif _HEX_STR_PATTERN6.match(colorstr):
            rgb = self._parse_hex(colorstr, 6)

        if rgb is None:
            raise FaviconColorError(colorstr)

        return rgb

    @staticmethod
    def _validate_rgb_iter(rgb_iter: Collection) -> Generator:
        """Validate & parse color iterable of RGB values."""

        if not len(rgb_iter) == 3:
            raise FaviconColorError(rgb_iter)

        for i in rgb_iter:
            try:
                num = int(i)
                if num not in range(0, 256):
                    raise FaviconColorError(rgb_iter)
                yield num
            except TypeError as err:
                raise FaviconColorError(rgb_iter) from err

    @staticmethod
    def _parse_rgbstr(rgbstr: str) -> Generator:
        """Validate & parse color string in rgb(x, y, z) format."""
        for digit in get_matches(_RGB_STR_PATTERN, rgbstr):
            num = int(digit)
            if num not in range(0, 256):
                raise FaviconColorError(rgbstr)
            yield num

    @staticmethod
    def _parse_hex(hexstr: str, length: int) -> Generator[int, None, None]:
        """Validate & parse color string in hex format."""
        rgb = None

        if length == 3:
            rgb = (s + s for s in get_matches(_HEX_STR_PATTERN3, hexstr))
        elif length == 6:
            rgb = get_matches(_HEX_STR_PATTERN6, hexstr)

        if rgb is None:
            raise FaviconColorError(hexstr)

        for c in rgb:
            yield int(c, 16)

    def __str__(self) -> str:
        """Represent instance as string."""
        return self.as_hex()

    def __repr__(self) -> str:
        """Represent instance & values."""
        colors = (f"{c}={getattr(self, c)}" for c in ("red", "green", "blue"))
        return f"{self.__class__.__name__}({', '.join(colors)})"



import logging
from pathlib import Path
from textwrap import indent
from typing import Any, Literal, overload

from hexdoc.core import (
    MinecraftVersion,
    ModResourceLoader,
    Properties,
    ResourceLocation,
)
from hexdoc.data import HexdocMetadata, load_metadata_textures
from hexdoc.minecraft import I18n, Tag
from hexdoc.minecraft.assets import (
    AnimatedTexture,
    HexdocAssetLoader,
    PNGTexture,
    Texture,
    TextureContext,
    TextureLookups,
)
from hexdoc.patchouli import BookContext
from hexdoc.patchouli.text import DEFAULT_MACROS, FormattingContext
from hexdoc.plugin import BookPlugin, ModPlugin, ModPluginWithBook, PluginManager

from .info import get_header

logger = logging.getLogger(__name__)


@overload
def load_common_data(
    props_file: Path,
    branch: str,
    *,
    book: Literal[True],
) -> tuple[Properties, PluginManager, BookPlugin[Any], ModPluginWithBook]: ...


@overload
def load_common_data(
    props_file: Path,
    branch: str,
    *,
    book: bool = False,
) -> tuple[Properties, PluginManager, BookPlugin[Any], ModPlugin]: ...


def load_common_data(
    props_file: Path,
    branch: str,
    *,
    book: bool = False,
) -> tuple[Properties, PluginManager, BookPlugin[Any], ModPlugin]:
    props = Properties.load(props_file)
    pm = PluginManager(branch, props)

    book_plugin = pm.book_plugin(props.book_type)
    mod_plugin = pm.mod_plugin(props.modid, book=book)

    header = get_header(props, pm, mod_plugin)
    logger.info(f"Loading hexdoc.\n{indent(header, '  ')}")

    minecraft_version = MinecraftVersion.MINECRAFT_VERSION = pm.minecraft_version()
    if minecraft_version is None:
        logger.warning(
            "No plugins implement minecraft_version. All versions may be used."
        )

    return props, pm, book_plugin, mod_plugin


def render_textures_and_export_metadata(
    loader: ModResourceLoader,
    asset_loader: HexdocAssetLoader,
):
    all_metadata = loader.load_metadata(model_type=HexdocMetadata)

    all_lookups = load_metadata_textures(all_metadata)
    image_textures = {
        id: texture
        for texture_type in [PNGTexture, AnimatedTexture]
        for id, texture in texture_type.get_lookup(all_lookups).items()
    }

    internal_lookups = TextureLookups[Texture](dict)
    if loader.props.textures.enabled:
        logger.info(f"Loading and rendering textures to {asset_loader.render_dir}.")
        for id, texture in asset_loader.load_and_render_internal_textures(
            image_textures
        ):
            texture.insert_texture(internal_lookups, id)

    # this mod's metadata
    metadata = HexdocMetadata(
        book_url=asset_loader.site_url / loader.props.default_lang,
        asset_url=loader.props.env.asset_url,
        textures=internal_lookups,
    )

    loader.export(
        metadata.path(loader.props.modid),
        metadata.model_dump_json(
            by_alias=True,
            warnings=False,
            exclude_defaults=True,
            round_trip=True,
        ),
        cache=True,
    )

    return all_metadata | {loader.props.modid: metadata}


# TODO: refactor a lot of this out
def init_context(
    *,
    book_id: ResourceLocation,
    book_data: dict[str, Any],
    pm: PluginManager,
    loader: ModResourceLoader,
    i18n: I18n,
    all_metadata: dict[str, HexdocMetadata],
):
    props = loader.props

    context = dict[str, Any]()

    for item in [
        props,
        pm,
        loader,
        i18n,
        TextureContext(
            textures=load_metadata_textures(all_metadata),
            allowed_missing_textures=props.textures.missing,
        ),
        FormattingContext(
            book_id=book_id,
            macros=DEFAULT_MACROS | book_data.get("macros", {}) | props.macros,
        ),
        BookContext(
            modid=props.modid,
            book_id=book_id,
            spoilered_advancements=Tag.SPOILERED_ADVANCEMENTS.load(
                loader
            ).value_ids_set,
            all_metadata=all_metadata,
            flags=pm.load_flags(),
        ),
    ]:
        item.add_to_context(context)

    for item in pm.update_context(context):
        item.add_to_context(context)

    return context



import logging
from typing import Any, TypeGuard, TypeVar, get_origin

logger = logging.getLogger(__name__)

_T = TypeVar("_T")

_DEFAULT_MESSAGE_SHORT = "Expected any of {expected}, got {actual}"
_DEFAULT_MESSAGE_LONG = "Expected any of {expected}, got {actual}: {value}"


def isinstance_or_raise(
    val: Any,
    class_or_tuple: type[_T] | tuple[type[_T], ...],
    message: str | None = None,
) -> TypeGuard[_T]:
    """Usage: `assert isinstance_or_raise(val, str)`

    message placeholders: `{expected}`, `{actual}`, `{value}`
    """

    # convert generic types into the origin type
    if not isinstance(class_or_tuple, tuple):
        class_or_tuple = (class_or_tuple,)
    ungenericed_classes = tuple(get_origin(t) or t for t in class_or_tuple)

    if not isinstance(val, ungenericed_classes):
        # just in case the caller messed up the message formatting
        subs = {
            "expected": list(class_or_tuple),
            "actual": type(val),
            "value": val,
        }

        if logger.getEffectiveLevel() >= logging.WARNING:
            default_message = _DEFAULT_MESSAGE_SHORT
        else:
            default_message = _DEFAULT_MESSAGE_LONG

        if message is None:
            raise TypeError(default_message.format(**subs))

        try:
            raise TypeError(message.format(**subs))
        except KeyError:
            raise TypeError(default_message.format(**subs))

    return True


def cast_or_raise(
    val: Any,
    class_or_tuple: type[_T] | tuple[type[_T], ...],
    message: str | None = None,
) -> _T:
    assert isinstance_or_raise(val, class_or_tuple, message)
    return val



from typing import Unpack

from hexdoc.core import ResourceLocation
from hexdoc.model import TypeTaggedTemplate
from hexdoc.utils import Inherit, InheritType, NoValue, classproperty
from pydantic import ConfigDict


class Page(TypeTaggedTemplate, type=None):
    """Base class for Patchouli page types.

    See: https://vazkiimods.github.io/Patchouli/docs/patchouli-basics/page-types
    """

    anchor: str | None = None

    def __init_subclass__(
        cls,
        *,
        type: str | InheritType | None = Inherit,
        template_type: str | None = None,
        **kwargs: Unpack[ConfigDict],
    ) -> None:
        super().__init_subclass__(type=type, template_type=template_type, **kwargs)

    @classproperty
    @classmethod
    def type(cls) -> ResourceLocation | None:
        assert cls._type is not NoValue
        return cls._type

    @classproperty
    @classmethod
    def template(cls) -> str:
        return f"pages/{cls.template_id.namespace}/{cls.template_id.path}"
